# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized two-dimensional convolution.

Quantized convolution abstracts over the analogous :py:func:`tf.conv2d` and
:py:func:`tf.conv2d_transpose` APIs to provide their quantized analogues.

Both the filter and input arguments to the convolution are quantized.
One difference in the interface for AQTp here compared to the Jax one
is that filter and input dilations are relegated to separate functions
(`conv2d` and `conv2d_transpose`), whereas in Jax the general convolution
API provides both dilations simultaneously. This reflects the underlying
frameworks' differences.

Further, note that we do not allow different weighting for kernel parameters,
unlike the general :py:func:`aqt_ops.matmul`.
"""

from aqt.common import aqt_config
from aqt.common import aqt_config_utils
from aqt.tensorflow import aqt_tensor
import tensorflow.compat.v1 as tf

# We repeatedly use protected methods from classes defined in other modules to
# avoid exporting them as part of the public API.
# pylint: disable=protected-access


def _assert_dilation_argument(lhs_quantizer, rhs_quantizer, dilations):
  """Validates preserve_zero=True of AqtIntQuantConfig when the dilation is used."""
  lhs_configs = lhs_quantizer.config.tensor_configs
  rhs_configs = rhs_quantizer.config.tensor_configs
  event_count = lhs_quantizer._last_update

  is_valid = tf.constant(True)
  for lhs_config, rhs_config in zip(lhs_configs, rhs_configs):
    if (isinstance(lhs_config.quant_config, aqt_config.IntQuantConfig) and
        isinstance(rhs_config.quant_config, aqt_config.IntQuantConfig) and
        lhs_config.quant_config.bits <= 8 and
        rhs_config.quant_config.bits <= 8):
      is_config_active = aqt_tensor.is_config_active(rhs_config, event_count)

      is_valid &= (~is_config_active | (dilations is None)
                   | rhs_config.quant_config.preserve_zero)

  assert_op = tf.debugging.Assert(is_valid, [
      'rhs_config.quant_config.preserve_zero must be True if the filter is dilated.'   # pylint: disable=line-too-long
  ])

  return assert_op


def _validate_contraction(
    config: aqt_config.StatsConfig,  # Prevent python auto-formatting.
    config_path: str,
    data_format: str) -> None:
  """Validates that conv2d contraction axes have shared stats.

  The convolution operation can be viewed as a higher-order tensor contraction
  along the height, width, and input channel axes for its input and filter.
  When quantizing the convolution, we thus have to keep the same multiplicative
  scale during calibration along these axes. Otherwise, we'd be changing the
  implicit inner product the operation computes in a manner besides
  quantization.

  Args:
    config: The configuration for recording conv2d argument statistics.
    config_path: A string for clearer error messages, to refer to the config.
    data_format: the format of the tensor, with H denoting height, W width, and
      C input channels.

  Raises:
    aqt_config.ConfigError: `config.share_stats_axes` does not include any
      of the contraction axes.
  """

  stats_axes = config.share_stats_axes

  axis_names = ['height', 'width', 'channel']
  dims = ['H', 'W', 'C']

  for axis_name, dim in zip(axis_names, dims):
    axis = data_format.index(dim)
    if axis not in stats_axes:
      raise aqt_config.ConfigError(
          f'expected {axis_name} contraction axis ({axis}) to be in '
          f'{config_path}.share_stats_axes={stats_axes}')


def _validate_inputs(
    input_quantizer: aqt_tensor.TensorQuantizer,  #
    filter_quantizer: aqt_tensor.TensorQuantizer,
    data_format: str) -> None:
  """Validates configs and inputs for conv2d.

  Args:
    input_quantizer: the input tensor quantizer.
    filter_quantizer: the filter tensor quantizer.
    data_format: the conv2d input tensor argument axes format.

  Raises:
    aqt_config.ConfigError: The input or filter quantizer configurations
      do not share statistics along contraction axes as described by
      `_validate_contraction`, or their quantization schedules are not aligned
      per `aqt_config_utils._validate_alignment`.
  """

  aqt_config_utils._validate_alignment(
      'input_quantizer.config.tensor_configs',
      input_quantizer.config.tensor_configs,
      'filter_quantizer.config.tensor_configs',
      filter_quantizer.config.tensor_configs)

  _validate_contraction(input_quantizer.config.stats_config,
                        'input_quantizer.config.stats_config', data_format)

  # Filters are always assumed to be in HWCO format, where
  # H - height, W - width, O - output channels, C - input channels.
  # Note input is NHWC or NCHW where N - batch size.
  _validate_contraction(filter_quantizer.config.stats_config,
                        'filter_quantizer.config.stats_config', 'HWCO')


def conv2d(
    input_quantizer: aqt_tensor.TensorQuantizer,  #
    input: tf.Tensor,  # pylint: disable=redefined-builtin
    filter_quantizer: aqt_tensor.TensorQuantizer,
    filter: tf.Tensor,  # pylint: disable=redefined-builtin
    train: bool = True,
    **tf_conv2d_kwargs):
  r"""Quantized :py:func:`tf.nn.conv2d`.

  Quantization is symmetric and uniform, with range determined by the
  `config`. Otherwise, all arguments are passed on to
  :py:func:`tf.nn.conv2d`.

  Args:
    input_quantizer: TensorQuantizer for input
    input: A `tf.Tensor`. Must be `float32`. The convolution input.
    filter_quantizer: TensorQuantizer for filter
    filter: A `tf.Tensor`. Must have the same type as `input`. The convolution
      kernel.
    train: if False, allows static switching for inference quantization
      configuration, a performance optimization.
    **tf_conv2d_kwargs: Keyword arguments to pass onto `conv2d`.

  Returns:
    A tensor of the same type as `input` conformal to what `conv2d` would
    return.
  """
  data_format = tf_conv2d_kwargs.get('data_format', 'NHWC')
  _validate_inputs(input_quantizer, filter_quantizer, data_format)
  assert_op = _assert_dilation_argument(input_quantizer, filter_quantizer,
                                        tf_conv2d_kwargs.get('dilations', None))

  with tf.control_dependencies([assert_op]):
    with tf.name_scope('AqtConv2d'):
      with tf.name_scope('get_quant_scale'):
        with tf.name_scope('input'):
          input_scale, input_inv_scale = input_quantizer._get_quant_scale(
              train)  # [N, 1, 1, 1]
        with tf.name_scope('filter'):
          filter_scale, filter_inv_scale = filter_quantizer._get_quant_scale(
              train)  # [1, 1, 1, O]

      input = input_scale * input  # [N, 1, 1, 1] * [N, H, W, C]
      filter = filter_scale * filter  # [1, 1, 1, O] * [H, W, I, O]

      with tf.name_scope('to_quant'):
        with tf.name_scope('input'):
          input = input_quantizer._to_quant(input, train)  # [N, H, W, C]
        with tf.name_scope('filter'):
          filter = filter_quantizer._to_quant(filter, train)  # [H, W, I, O]

      # TODO(vladf): until tf.conv2d supports int8 arguments, we need to cast
      # the quantized variables to a floating point format.
      if input_quantizer.config.use_quantized_variable and not train:
        input = tf.cast(input_quantizer.quantized_variable.read_value(),
                        tf.float32)
      if filter_quantizer.config.use_quantized_variable and not train:
        filter = tf.cast(filter_quantizer.quantized_variable.read_value(),
                         tf.float32)

      with tf.name_scope('conv2d'):
        # TODO(vladf): implement calls to narrowed functions, e.g., int8_conv2d,
        # cased by the bitwidth settings for our tensor quantization
        # configurations, in preparation for quantization support in conv2d.
        conv = tf.nn.conv2d(input, filter, **tf_conv2d_kwargs)  # [N, H, W, O]

      with tf.name_scope('inv_scale'):
        # [N, H, W, O] * ([N, 1, 1, 1] * [1, 1, 1, O])
        return conv * (input_inv_scale * filter_inv_scale)


def depthwise_conv2d(
    input_quantizer: aqt_tensor.TensorQuantizer,  #
    input: tf.Tensor,  # pylint: disable=redefined-builtin
    filter_quantizer: aqt_tensor.TensorQuantizer,
    filter: tf.Tensor,  # pylint: disable=redefined-builtin
    train: bool = True,
    **tf_dw_conv2d_kwargs):
  r"""Quantized :py:func:`tf.nn.depthwise_conv2d`.

  Quantization is symmetric and uniform, with range determined by the
  `config`. Otherwise, all arguments are passed on to
  :py:func:`tf.nn.depthwise_conv2d`.

  Args:
    input_quantizer: TensorQuantizer for input
    input: A `tf.Tensor`. Must be `float32`. The convolution input.
    filter_quantizer: TensorQuantizer for filter
    filter: A `tf.Tensor`. Must have the same type as `input`. The convolution
      kernel.
    train: if False, allows static switching for inference quantization
      configuration, a performance optimization.
    **tf_dw_conv2d_kwargs: Keyword arguments to pass onto `depthwise_conv2d`.

  Returns:
    A tensor of the same type as `input` conformal to what
    `depthwise_conv2d` would return.
  """
  data_format = tf_dw_conv2d_kwargs.get('data_format', 'NHWC')
  _validate_inputs(input_quantizer, filter_quantizer, data_format)
  assert_op = _assert_dilation_argument(
      input_quantizer, filter_quantizer,
      tf_dw_conv2d_kwargs.get('dilations', None))

  with tf.control_dependencies([assert_op]):
    with tf.name_scope('AqtDepthwiseConv2d'):
      with tf.name_scope('get_quant_scale'):
        with tf.name_scope('input'):
          input_scale, input_inv_scale = input_quantizer._get_quant_scale(
              train)  # [N, 1, 1, 1]
        with tf.name_scope('filter'):
          filter_scale, filter_inv_scale = filter_quantizer._get_quant_scale(
              train)  # [1, 1, 1, O]

      input = input_scale * input  # [N, 1, 1, 1] * [N, H, W, C]
      filter = filter_scale * filter  # [1, 1, 1, O] * [H, W, I, O]

      with tf.name_scope('to_quant'):
        with tf.name_scope('input'):
          input = input_quantizer._to_quant(input, train)  # [N, H, W, C]
        with tf.name_scope('filter'):
          filter = filter_quantizer._to_quant(filter, train)  # [H, W, I, O]

      if input_quantizer.config.use_quantized_variable and not train:
        input = tf.cast(input_quantizer.quantized_variable.read_value(),
                        tf.float32)
      if filter_quantizer.config.use_quantized_variable and not train:
        filter = tf.cast(filter_quantizer.quantized_variable.read_value(),
                         tf.float32)

      with tf.name_scope('depthwise_conv2d'):
        dw_conv = tf.nn.depthwise_conv2d(input, filter, **tf_dw_conv2d_kwargs)

      with tf.name_scope('inv_scale'):
        # [N, H, W, O] * ([N, 1, 1, 1] * [1, 1, 1, O])
        return dw_conv * (input_inv_scale * filter_inv_scale)



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for conv2d."""

from absl.testing import absltest
from aqt.tensorflow import aqt_ops
from aqt.tensorflow import aqt_tensor
from aqt.test import aqt_conv_test_base
import tensorflow.compat.v1 as tf


class Conv2dTest(aqt_conv_test_base.ConvTest):

  def conv_op_quantized(
      self,
      input,  # pylint: disable=redefined-builtin
      filter,  # pylint: disable=redefined-builtin
      input_config,
      filter_config,
      event_count,
      event_count_for_filter=None,
      input_weights=None,
      train=True,
      var_scope_name=None,
      **kwargs):
    event_count_for_filter = tf.constant(
        event_count_for_filter if event_count_for_filter else event_count,
        dtype=tf.int64)
    event_count = tf.constant(event_count, dtype=tf.int64)

    input = self.constant(input)
    filter = self.constant(filter)

    with tf.variable_scope(var_scope_name, default_name="default"):
      input_tq = aqt_tensor.TensorQuantizer(
          input.shape, input_config, name="input")
      filter_tq = aqt_tensor.TensorQuantizer(
          filter.shape, filter_config, name="filter")

    updates = [input_tq.update(input, input_weights, event_count),
               filter_tq.update(filter, None, event_count_for_filter)]

    with tf.control_dependencies(updates):
      result = aqt_ops.aqt_conv2d(input_tq, input, filter_tq, filter, **kwargs)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      return result.eval()

  def conv_op_unquantized(self, input, filter, **kwargs):  # pylint: disable=redefined-builtin
    input = self.constant(input)
    filter = self.constant(filter)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      return tf.nn.conv2d(input, filter, **kwargs).eval()

  def get_conv_kwargs(
      self,
      strides,
      padding,
      data_format="NHWC",
      dilations=None):
    return dict(
        strides=strides,
        padding=padding,
        data_format=data_format,
        dilations=dilations)

  def constant(self, x):
    return tf.constant(x)

  def gradients(self, fwd_func, x, w):
    fwd = fwd_func(x, w)
    return tf.gradients([fwd], [x, w])

  def test_vars_over_inputs_at_inference(self):
    input_config, x, filter_config, w = self.exact_int8_conv_example(
        lhs_use_quantized_variable=True, rhs_use_quantized_variable=True)

    x = self.constant(x)
    w = self.constant(w)
    kwargs = self.get_conv_kwargs(strides=1, padding="VALID")

    input_tq = aqt_tensor.TensorQuantizer(x.shape, input_config, name="input")
    filter_tq = aqt_tensor.TensorQuantizer(
        w.shape, filter_config, name="filter")
    event_count = tf.constant(0, dtype=tf.int64)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()

      input_tq.update(x, None, event_count).run()
      filter_tq.update(w, None, event_count).run()

      # Since quantized variables are not used at training, the conv2d with zero
      # inputs should produce zero values.
      actual_train = aqt_ops.aqt_conv2d(
          input_tq,
          tf.zeros_like(x),
          filter_tq,
          tf.zeros_like(w),
          train=True,
          **kwargs)
      expected = tf.zeros_like(actual_train)
      self.assertAllEqual(actual_train, expected)

      # Since quantized variables should be always used at inference, the conv2d
      # will rely on quantized variables.
      actual_infer = aqt_ops.aqt_conv2d(
          input_tq,
          tf.zeros_like(x),
          filter_tq,
          tf.zeros_like(w),
          train=False,
          **kwargs)
      expected = aqt_ops.aqt_conv2d(
          input_tq, x, filter_tq, w, train=True, **kwargs)
      self.assertAllEqual(actual_infer, expected)


class DepthwiseConv2dTest(Conv2dTest):

  def conv_op_quantized(
      self,
      input,  # pylint: disable=redefined-builtin
      filter,  # pylint: disable=redefined-builtin
      input_config,
      filter_config,
      event_count,
      event_count_for_filter=None,
      input_weights=None,
      train=True,
      var_scope_name=None,
      **kwargs):
    event_count_for_filter = tf.constant(
        event_count_for_filter if event_count_for_filter else event_count,
        dtype=tf.int64)
    event_count = tf.constant(event_count, dtype=tf.int64)

    input = self.constant(input)
    filter = self.constant(filter)

    with tf.variable_scope(var_scope_name, default_name="default"):
      input_tq = aqt_tensor.TensorQuantizer(
          input.shape, input_config, name="input")
      filter_tq = aqt_tensor.TensorQuantizer(
          filter.shape, filter_config, name="filter")

    updates = [input_tq.update(input, input_weights, event_count),
               filter_tq.update(filter, None, event_count_for_filter)]

    # depthwise_conv2d expects a different stride format than conv2d
    if "strides" in kwargs:
      stride = kwargs["strides"]
      kwargs["strides"] = [1, stride, stride, 1]

    with tf.control_dependencies(updates):
      result = aqt_ops.aqt_depthwise_conv2d(input_tq, input, filter_tq,
                                            filter, **kwargs)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      return result.eval()

  def conv_op_unquantized(self, input, filter, **kwargs):  # pylint: disable=redefined-builtin
    input = self.constant(input)
    filter = self.constant(filter)

    # tf.nn.depthwise_conv2d expects a different stride format than *_conv2d
    if "strides" in kwargs:
      stride = kwargs["strides"]
      kwargs["strides"] = [1, stride, stride, 1]

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      return tf.nn.depthwise_conv2d(input, filter, **kwargs).eval()

if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common utility functions for AQT ops.
"""

from typing import Callable, Dict, List, Optional

from aqt.common import aqt_config
from aqt.tensorflow import aqt_tensor
import tensorflow.compat.v1 as tf

# TODO(b/220181240): Remove the pylint disable below and avoid using protected
# methods.
# We repeatedly use protected methods from classes defined in other modules to
# avoid exporting them as part of the public API.
# pylint: disable=protected-access

TensorQuantizer = aqt_tensor.TensorQuantizer
DynamicTensorQuantizer = aqt_tensor.DynamicTensorQuantizer


def _possibly_use_quantized_variable(
    quantizer: TensorQuantizer, x: tf.Tensor, train: bool  #
) -> tf.Tensor:
  """Returns quantized variable if not training and TQ.use_quantized_variable, casted to x.dtype.

  Given an input tensor and its tensor quantizer, here we enforce to use the
  quantized variable stored in the tensor quantizer as long as
  TQ.use_quantized_variable is true and it is in inference, no matter if
  FloatConfig is specified or not.

  The semantics of FloatConfig which is meant not to use quantized variables
  during inference should be respected by requiring users to specify
  TensorQuantizer.use_quantized_variable=False. See more details at
  b/219040448.

  Args:
    quantizer: TensorQuantizer for the input tensor x.
    x: lhs or rhs of matmul or einsum.
    train: Indicates if in training or not.

  Returns:
    The input tensor x or its quantized one.
  """
  if quantizer.config.use_quantized_variable and not train:
    qx = quantizer.quantized_variable
    if qx.dtype != x.dtype:
      qx = tf.cast(qx, x.dtype)
    return qx
  return x


def diagnostics(
    op,
    lhs: tf.Tensor,
    rhs: tf.Tensor,
    grad: Optional[tf.Tensor] = None,
) -> Dict[str, tf.Tensor]:
  """Returns a dictionary from keys to diagnostic tensors.

  Args:
    op: A quantized binary op.
    lhs: lhs argument to op.Apply, used for deriving diangostics relative to
      a given input.
    rhs: as above, but for rhs
    grad: If specified, the gradient for deriving diagnostics.

  Returns:
    A dictionary with various quantization-related diagnostics,
    whose string keys are prefixed by op.name/op.{lhs,rhs}_name.
  """
  d = {}
  quantizers = [
      (op.lhs_name, op.lhs_quantizer, lhs),
      (op.rhs_name, op.rhs_quantizer, rhs),
  ]
  if grad is not None:
    assert op.grad_quantizer is not None, (
        'If grad is given, then grad_quantizer must be defined.')
    quantizers.append((op.grad_name, op.grad_quantizer, grad))
  for prefix, quantizer, argument in quantizers:
    clipped_proportion = tf.cast(tf.abs(argument) > quantizer.clip_range(),
                                 tf.float32)
    prefix = f'{op.name}/{prefix}'
    d[f'{prefix}/clipped_proportion'] = tf.math.reduce_mean(
        clipped_proportion)
    d[f'{prefix}/clip'] = quantizer.clip_range()
    d[f'{prefix}/event_count'] = quantizer._last_update
    for name, var in quantizer.calibration_variables().items():
      d[f'{prefix}/{name}'] = var
  return d


def _dense_op_case(
    lhs_quantizer: TensorQuantizer | DynamicTensorQuantizer,
    rhs_quantizer: TensorQuantizer,
    default_dense_op: Callable[[], tf.Tensor],
    int8_dense_op: Callable[[], tf.Tensor],
    train: bool,
    skip_cond_if_all_int_quant: bool = False,
) -> tf.Tensor:
  """Switch over the dense compute based on event count and configs.

  The `TensorQuantizer`s for each argument are provided to supply the
  configurations for each input tensor. Also, if indicated by constituent
  configs, this uses the quantized variables in each tensor quantizer rather
  than the inputs themselves. Regardless, gradients are preserved as if this was
  a matmul over the original `(lhs, rhs)` inputs.

  Args:
    lhs_quantizer: TensorQuantizer for lhs.
    rhs_quantizer: TensorQuantizer for rhs.
    default_dense_op: function for the float dense op.
    int8_dense_op: function for the int8 dense op.
    train: If false and `TQ.use_quantized_variable` is True, then use the
      quantized variable, instead of input tensors, for the respective input
      tensor.
    skip_cond_if_all_int_quant: If all quant configs are IntQuantConfig, skip
      tf.cond.

  Returns:
    The `tf.Tensor` from the resulting quantized dense op.
  """

  lhs_configs = lhs_quantizer.config.tensor_configs
  rhs_configs = rhs_quantizer.config.tensor_configs

  def is_int8_compatible(
      lhs_config: aqt_config.AqtTensorConfig,
      rhs_config: aqt_config.AqtTensorConfig,
  ) -> bool:
    return (
        isinstance(lhs_config.quant_config, aqt_config.IntQuantConfig)
        and isinstance(rhs_config.quant_config, aqt_config.IntQuantConfig)
        and lhs_config.quant_config.bits <= 8
        and rhs_config.quant_config.bits <= 8
    )

  def all_int8_compatible(
      lhs_configs: List[aqt_config.AqtTensorConfig],
      rhs_configs: List[aqt_config.AqtTensorConfig],
  ) -> bool:
    for lhs_config in lhs_configs:
      for rhs_config in rhs_configs:
        if not is_int8_compatible(lhs_config, rhs_config):
          return False
    return True

  # short circuit for quantization all the way, which is not enabled by default
  # for backward checkpoint compatibility.
  if skip_cond_if_all_int_quant and all_int8_compatible(
      lhs_configs, rhs_configs
  ):
    return int8_dense_op()

  lhs_index = lhs_quantizer.config.inference_config_index
  rhs_index = rhs_quantizer.config.inference_config_index

  if train or lhs_index is None or rhs_index is None:
    should_int8_quantize = tf.constant(False)
    for lhs_config in lhs_configs:
      for rhs_config in rhs_configs:
        # If any of lhs and rhs is FloatConfig, use the default matmul.
        if is_int8_compatible(lhs_config, rhs_config):
          should_int8_quantize |= aqt_tensor.is_config_active(
              lhs_config, lhs_quantizer._last_update
          ) & aqt_tensor.is_config_active(
              rhs_config, rhs_quantizer._last_update
          )
    # Use go/tf-control-flow-v2, which we've noticed fuses better on TPU XLA.
    v2_was_enabled = tf.control_flow_v2_enabled()
    if not v2_was_enabled:
      tf.enable_control_flow_v2()
    cond = tf.cond(should_int8_quantize, int8_dense_op, default_dense_op)
    if not v2_was_enabled:
      tf.disable_control_flow_v2()
  else:
    # In the inference setting, if inference config indices are specified,
    # then manually const-prop the tf.cond to avoid overheads such as loading
    # bf16 weights
    should_int8_quantize = is_int8_compatible(
        lhs_configs[lhs_index], rhs_configs[rhs_index]
    )
    if should_int8_quantize:
      cond = int8_dense_op()
    else:
      cond = default_dense_op()
  return cond



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Accurate Quantized Training ops.

A core of the quantization algorithm is replacing float matmul:

r = matmul(a, b)

with a following equation:

sa = calibrate(a)
sb = calibrate(b)
sa1 = 1 / sa
sb1 = 1 / sb

def q(x):
  return round(clip(x, -clip_bound, clip_bound))

qa = q(sa * a)  # each row can be scaled separately
qb = q(b * sb)  # each column can be scaled separately
qr = int_matmul(qa, qb)
r = sa1 * qr * sb1

We can observe that if 'q' function was identity, both computations
would be equivalent. A similar approach applies to most linear operations.

Calibration is a major difficulty, i.e. how to find best scales sa, sb.
This is solved (or rather tackled) by Stats.bound() and StatsConfig.

Another difficulty is feedback loops between Stats.bound() calculation and
the changing distribution of inputs.
This can be tackled by setting freeze_scale_at_begin=True

Another difficulty is 'sparse' distributions.
E.g. quantization of the output of Relu and other similar functions.
This can be tackled with filter_zeros=True setting.

Typical quantization uses mean-of-batch-max calibration.
It is problematic as that statistics is dependant on batch size and can't
take example weights into account.
lp_dev with large lp_order is a suitable replacement.

Typical matmuls have weights that can have per-output-channel scales.
In the above equation 'b' would be weights and 'sb' diagonal matrix.
While activations usually share distribution across examples in batch and
may have 'sa' as a scalar. This can be configured using 'share_stats_axes'.

Note that share_stats_axes should always include contraction axis.
For instance in case of MatMul, share_stats_axes in lhs should include '1', and
share_stats_axes in rhs should include '0'.

There are several reasons to control schedule of quantization:

  - Avoid quantization of initial turbulent period of training to
    accelerate convergence.
  - Separately quantize weights and activations in various layers.
  - One also may wish to incrementally increase quantization.
  - Disable quantization just before the training is finished.
  - Try multiple configurations in one long 'anytime' training.

All and any of the above can be achieved by providing multiple
lhs_configs and rhs_configs with begin_at_event and end_at_event set properly.
The intervals [begin_at_event, end_at_event) are expected to be disjoint.
"""

from aqt.tensorflow import aqt_conv2d
from aqt.tensorflow import aqt_einsum
from aqt.tensorflow import aqt_matmul

# TODO(b/193796715): Consider using Union[tf.Tensor, tf.Variable].
# TODO(vladf): aqt_ops.aqt_matmul is redundant, we should consider
# renaming to just the short operation name, e.g., aqt_ops.matmul.

aqt_depthwise_conv2d = aqt_conv2d.depthwise_conv2d
aqt_conv2d = aqt_conv2d.conv2d
aqt_einsum = aqt_einsum.einsum
aqt_matmul = aqt_matmul.matmul



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for aqt_tensor."""

from typing import Set

from absl import logging
from absl.testing import absltest
from absl.testing import parameterized
from aqt.common import aqt_config
from aqt.tensorflow import aqt_tensor
from aqt.test import aqt_stats_test_base
from aqt.test import aqt_tensor_test_base
import numpy as np
import tensorflow.compat.v1 as tf


def f32(x):
  """Cast input array to f32."""
  return tf.cast(x, dtype=tf.float32)


def i64(x):
  """Cast input to i64."""
  return tf.cast(x, dtype=tf.int64)


def padded_weight(data_shape):
  rweight = np.random.uniform(low=0.0, high=1.0, size=data_shape)
  return np.where(rweight > 0.5, rweight, np.zeros_like(rweight))


none_weight = lambda _: None


class StatsTest(aqt_stats_test_base.StatsTest):
  """Tests for Stats class.

  Refer to aqt_stats_test_base.StatsTest for more details.
  """

  def set_stats(self, data_shape, config):
    self._stats = aqt_tensor.Stats(
        data_shape=data_shape,
        config=config,
        get_variable=aqt_tensor.default_get_variable)
    tf.global_variables_initializer().run()

  def update(self, sample, weight):
    self._stats.update(f32(sample), f32(weight)).run()

  def get_sum_of_ones(self):
    return self.stats.calibration_variables()['sum_of_ones']

  def get_sum_of_vals(self):
    return self.stats.calibration_variables()['sum_of_vals']

  def get_max_of_abs_vals(self):
    with self.cached_session() as sess, sess.as_default():
      return self.stats.max_dev().eval()

  def get_sum_of_l1_vals(self):
    return self.stats.calibration_variables()['sum_of_l1_vals']

  def get_sum_of_lp_vals(self):
    return self.stats.calibration_variables()['sum_of_lp_vals']

  def set_ema_update_count(self, ema_update_count):
    self.stats._ema_update_count = ema_update_count

  def mean(self):
    return self.stats.mean()

  def max_dev(self):
    return self.stats.max_dev()

  def l1_dev(self):
    return self.stats.l1_dev()

  def lp_dev(self):
    return self.stats.lp_dev()

  def bound(self, calibration_config):
    return self.stats.bound(calibration_config)


class AqtTensorQuantizerTest(aqt_tensor_test_base.AqtTensorQuantizerTest):
  """Tests for AqtTensorQuantizer class.

  Refer to aqt_tensor_test_base.AqtTensorQuantizerTest for more details.
  """

  def make_tensor_quantizer(
      self,
      data_shape,
      config,
      name='tq',
      quantizer_class=aqt_tensor.TensorQuantizer,
  ):
    return quantizer_class(
        data_shape=data_shape,
        config=config,
        name=name,
    )

  def update_quantizer(self, quant, sample, weight, event_count):
    with self.cached_session() as sess, sess.as_default():
      quant.update(f32(sample), weight, i64(event_count)).run()

  def to_quant(self, quant, x, train=True):
    with self.cached_session() as sess, sess.as_default():
      return quant._to_quant(x, train).eval()

  def get_quant_scale(self, quant, train=True):
    with self.cached_session() as sess, sess.as_default():
      scale, inv_scale = quant._get_quant_scale(train)
      return scale.eval(), inv_scale.eval()

  def get_dynamic_quant_scale(
      self, quant, sample, weight, event_count, train,
      data_shape,
  ):
    @tf.function(
        input_signature=[
            tf.TensorSpec(shape=data_shape, dtype=sample.dtype),
            ]
    )
    def _get_dynamic_quant_scale(sample):
      return quant._get_dynamic_quant_scale(
          sample, weight, event_count, train
      )
    with self.cached_session() as sess, sess.as_default():
      scale, inv_scale = _get_dynamic_quant_scale(sample)
      return scale.eval(), inv_scale.eval()

  def get_dynamic_clip_range(
      self, quant, sample, weight, event_count, train,
      data_shape,
  ):
    @tf.function(
        input_signature=[
            tf.TensorSpec(shape=data_shape, dtype=sample.dtype),
            ]
    )
    def _get_dynamic_clip_range(sample):
      return quant.dynamic_clip_range(
          sample, weight, event_count, train
      )
    with self.cached_session() as sess, sess.as_default():
      return _get_dynamic_clip_range(sample).eval()

  def init(self):
    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()

  def get_scale(self, quant):
    with self.cached_session() as sess, sess.as_default():
      return quant._scale.eval()

  def get_clip_range(self, quant):
    with self.cached_session() as sess, sess.as_default():
      return quant.clip_range().eval()

  def get_last_update(self, quant):
    with self.cached_session() as sess, sess.as_default():
      return quant._last_update.eval()

  def get_quantized_variable(self, quant):
    with self.cached_session() as sess, sess.as_default():
      return quant.quantized_variable.eval()

  def test_pure_quant_methods(self):
    """Validates the to_quant methods are pure functions of scale and input."""
    # Check the generated tensorflow graph to make sure that statistics are not
    # used and that the output is deterministic.

    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False,
    )
    config = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.IntQuantConfig(bits=8),
        calibration_config=aqt_config.CalibrationConfig(const_bound_coeff=1),
        freeze_scale_at_begin=False)
    config = aqt_config.AqtScheduleConfig(sc, [config])

    rng = np.random.default_rng(1234)
    x = rng.integers(-10, 10, size=(4, 4)).astype(np.float32)
    with tf.Graph().as_default():
      quant = aqt_tensor.TensorQuantizer(data_shape=x.shape, config=config)
      event_count = tf.constant(0, dtype=tf.int64)
      with self.cached_session():
        tf.global_variables_initializer().run()
        quant.update(f32(x), None, event_count).run()

      repeats = []
      for _ in range(10):
        scale, inv_scale = quant._get_quant_scale(train=True)
        x = scale * x
        ix = quant._to_quant(x, train=True)
        qx = inv_scale * ix
        repeats.append(qx)

      with self.cached_session():
        for l, r in zip(repeats, repeats[1:]):
          self.assertAllEqual(l, r)

    # Extract subgraph attached to `qx`.
    actual_varnames = extract_referenced_variables(qx)
    self.assertLen(actual_varnames, 2)
    expected_varnames = {quant._inv_scale.name, quant._scale.name}
    self.assertEqual(actual_varnames, expected_varnames)

  def test_alternative_variable_creation(self):
    """Makes sure that tensor quantization works with different GetVariable."""
    sc = aqt_config.StatsConfig(
        ema_update_count=1,
        share_stats_axes=[0, 1],
        tpu_cross_replica_sum=False,
    )
    config = aqt_config.AqtTensorConfig(
        aqt_config.IntQuantConfig(bits=8,),
        aqt_config.CalibrationConfig(l1_dev_coeff=2),
        freeze_scale_at_begin=False)
    config = aqt_config.AqtScheduleConfig(sc, [config])

    def alt_get_var(name, shape, dtype, init):
      init_fn = lambda: tf.constant(init, shape=shape, dtype=dtype)
      return tf.Variable(
          initial_value=init_fn,
          trainable=False,
          shape=shape,
          name=name,
          use_resource=True)

    rng = np.random.default_rng(1234)
    x = rng.integers(-10, 10, size=(4, 4)).astype(np.float32)
    with tf.Graph().as_default():
      var_quant = aqt_tensor.TensorQuantizer(
          x.shape, config, name='default_vars')
      alt_quant = aqt_tensor.TensorQuantizer(
          x.shape, config, name='new_vars', get_variable=alt_get_var)

      var_update, alt_update = tf.no_op(), tf.no_op()

      for i in range(1, 5):
        # Simulate multiple updates in a row.
        event_count = tf.constant(0, dtype=tf.int64)
        y = x * i / 4

        with tf.control_dependencies([var_update]):
          var_update = var_quant.update(f32(y), None, event_count)
        with tf.control_dependencies([alt_update]):
          alt_update = alt_quant.update(f32(y), None, event_count)

        with tf.control_dependencies([var_update]):
          scale, inv_scale = var_quant._get_quant_scale(train=True)
          y_scaled = scale * y
          ix = var_quant._to_quant(y_scaled, train=True)
          var_qx = inv_scale * ix

        with tf.control_dependencies([alt_update]):
          scale, inv_scale = alt_quant._get_quant_scale(train=True)
          y_scaled = scale * y
          ix = alt_quant._to_quant(y_scaled, train=True)
          alt_qx = inv_scale * ix

        with self.cached_session():
          tf.global_variables_initializer().run()
          self.assertAllEqual(var_qx, alt_qx)

  @parameterized.product(
      get_weight=[
          none_weight,
          padded_weight,
      ],
      calibration_config=[
          aqt_config.CalibrationConfig(
              max_dev_coeff=1,
          ),
          aqt_config.CalibrationConfig(
              const_bound_coeff=2,
          ),
          aqt_config.CalibrationConfig(
              l1_dev_coeff=3,
          ),
          aqt_config.CalibrationConfig(
              lp_dev_coeff=4,
          ),
      ],
  )
  def test_dynamic_quant_consistent(self, get_weight, calibration_config):
    """Tests basic dynamic quantization behavior is consistent whether toggle dynamic."""
    bits = 8
    x_bound = 16.0

    sc = aqt_config.StatsConfig(
        ema_update_count=1, share_stats_axes=[1], tpu_cross_replica_sum=False
    )
    tensor_config = aqt_config.AqtTensorConfig(
        quant_config=aqt_config.IntQuantConfig(bits),
        freeze_scale_at_begin=False,
        calibration_config=calibration_config,
        begin_at_event=10,
        end_at_event=20,
    )
    config = aqt_config.AqtScheduleConfig(sc, [tensor_config])

    data_shape = [128, 256]
    # replace a non-shared stats axis to None to test dynamic shape
    dynamic_data_shape = [None, data_shape[1]]
    assert sc.share_stats_axes == [1]

    weight = get_weight(data_shape)
    if weight is not None:
      weight = f32(weight)

    quant = self.make_tensor_quantizer(data_shape, config, 'tq')
    dyn_quant = self.make_tensor_quantizer(
        dynamic_data_shape,
        config,
        'dtq',
        quantizer_class=aqt_tensor.DynamicTensorQuantizer,
    )
    self.init()
    for new_event_count, should_quantize in [
        (9, False),
        (10, True),
        (19, True),
        (20, False),
    ]:
      logging.info('loop: %s', (new_event_count, should_quantize))
      x_init = np.random.uniform(low=-x_bound, high=x_bound, size=data_shape)
      x_with_zeros = np.where(x_init > 0.5, x_init, np.zeros_like(x_init))
      x = f32(x_with_zeros)
      new_event_count = np.array(new_event_count, dtype=np.int64)
      self.update_quantizer(quant, x, weight, new_event_count)

      self.assertAllEqual(self.get_last_update(quant), new_event_count)

      scale, inv_scale = self.get_quant_scale(quant)
      clip_range = self.get_clip_range(quant)
      ix = self.to_quant(quant, scale * x)

      dyn_scale, dyn_inv_scale = self.get_dynamic_quant_scale(
          dyn_quant, x, weight, new_event_count, train=True,
          data_shape=dynamic_data_shape,
      )
      dyn_clip_range = self.get_dynamic_clip_range(
          dyn_quant, x, weight, new_event_count, train=True,
          data_shape=dynamic_data_shape
      )
      dyn_ix = self.to_quant(dyn_quant, dyn_scale * x)

      with self.subTest('scale'):
        self.assertAllEqual(scale, dyn_scale)
      with self.subTest('inv_scale'):
        self.assertAllEqual(inv_scale, dyn_inv_scale)
      with self.subTest('clip_range'):
        self.assertAllEqual(clip_range, dyn_clip_range)
      with self.subTest('to_quant'):
        self.assertAllEqual(ix, dyn_ix)


def extract_referenced_variables(t: tf.Tensor) -> Set[str]:
  """Returns a set of all referenced variable names in a tensor's graph."""

  tensors = {t}

  def dfs(op):
    for tensor in op.inputs:
      if tensor not in tensors:
        tensors.add(tensor)
        dfs(tensor.op)
    for control_op in op.control_inputs:
      dfs(control_op)

  dfs(t.op)

  variables = [
      v for v in tensors
      if v.op.type == 'VarHandleOp' or v.op.type == 'VariableV2'
  ]

  return {v.name for v in variables}


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for utility function."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from aqt.tensorflow import aqt_ops_util
from aqt.tensorflow import aqt_tensor
from aqt.test import aqt_matmul_test_base
import tensorflow.compat.v1 as tf


class IntNarrowedMatMulTest(tf.test.TestCase, parameterized.TestCase):

  def test_chooses_right_matmul_for_schedule(self):
    # Create a list of settings (left_bits, right_bits, expected_matmul)
    # and generate a schedule based on those settings.
    settings = [(8, 16, "default"), (4, 16, "default"), (4, 8, "int8"),
                (8, 9, "default"), (4, 4, "int8"), (3, 3, "int8"),
                (9, 7, "default")]

    lhs_schedule = []
    rhs_schedule = []
    expected_results = []
    for i, (l, r, expected) in enumerate(settings):
      lhs_schedule.append((i, i + 1, l))
      rhs_schedule.append((i, i + 1, r))
      expected_results.append(expected)

    # any config that does not raise an error
    lhs_config = aqt_matmul_test_base.config_from_schedule(lhs_schedule)
    rhs_config = aqt_matmul_test_base.config_from_schedule(rhs_schedule)

    shape = [1, 1]  # Any shape will do, we're mocking.
    lhs_quant = aqt_tensor.TensorQuantizer(shape, lhs_config, name="lhs")
    rhs_quant = aqt_tensor.TensorQuantizer(shape, rhs_config, name="rhs")

    default_fn = mock.Mock(return_value=tf.constant("default"))
    int8_fn = mock.Mock(return_value=tf.constant("int8"))

    event_ph = tf.placeholder(tf.int64)
    lhs_quant._last_update = event_ph
    rhs_quant._last_update = event_ph
    train = True
    tf_actual = aqt_ops_util._dense_op_case(
        lhs_quant, rhs_quant, default_fn, int8_fn, train
    )

    with self.cached_session():
      tf.global_variables_initializer().run()
      for i, expected in enumerate(expected_results):
        actual = tf_actual.eval(feed_dict={event_ph: i})
        self.assertEqual(
            actual.decode("utf-8"), expected, msg=f"event_count {i}"
        )

  def test_quant_all_the_time(self):
    # Create a list of settings (left_bits, right_bits, expected_matmul)
    # and generate a schedule based on those settings.
    settings = [(8, 8, "int8"), (8, 8, "int8")]

    lhs_schedule = []
    rhs_schedule = []
    expected_results = []
    for i, (l, r, expected) in enumerate(settings):
      lhs_schedule.append((i, i + 1, l))
      rhs_schedule.append((i, i + 1, r))
      expected_results.append(expected)

    # any config that does not raise an error
    lhs_config = aqt_matmul_test_base.config_from_schedule(lhs_schedule)
    rhs_config = aqt_matmul_test_base.config_from_schedule(rhs_schedule)

    shape = [1, 1]  # Any shape will do, we're mocking.
    lhs_quant = aqt_tensor.TensorQuantizer(shape, lhs_config, name="lhs")
    rhs_quant = aqt_tensor.TensorQuantizer(shape, rhs_config, name="rhs")

    default_fn = mock.Mock(return_value=tf.constant("default"))
    int8_fn = mock.Mock(return_value=tf.constant("int8"))

    event_ph = tf.placeholder(tf.int64)
    lhs_quant._last_update = event_ph
    rhs_quant._last_update = event_ph
    train = True
    tf_actual = aqt_ops_util._dense_op_case(
        lhs_quant, rhs_quant, default_fn, int8_fn, train
    )

    with self.cached_session():
      tf.global_variables_initializer().run()
      for i, expected in enumerate(expected_results):
        actual = tf_actual.eval(feed_dict={event_ph: i})
        self.assertEqual(
            actual.decode("utf-8"), expected, msg=f"event_count {i}"
        )


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Small-sized model with 2-bit quantization for dense/attention weights only."""

from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs import base_config
from aqt.jax_legacy.jax.wmt_mlperf.hparams_configs.experimental import small_model_bfloat16


def get_config():
  """Returns configuration for a small transformer model with 2-bit quantization for dense/attention weights only."""
  config = small_model_bfloat16.get_config(
      quant_target=base_config.QuantTarget.WEIGHTS_ONLY)
  config.quant_type = "aqt"
  config.mlp_block.dense_1.weight_prec = 2
  config.mlp_block.dense_2.weight_prec = 2
  config.attention.dense_kqv.weight_prec = 2
  config.attention.dense_out.weight_prec = 2
  config.metadata.hyper_str = "mlp_attn_weights_quant_2bit"
  return config



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized dot_general."""

# Lingo in this file:
#
# - lhs(rhs) - left(right) hand side of a binary operation
# - ca - contraction axes
# - ba - batch axes
# - ra - remaining axes

# pylint: disable=g-explicit-bool-comparison
# pylint: disable=g-explicit-length-test

import copy
import functools
from typing import Callable, Optional, Union
from aqt.jax.v2 import config
import flax.struct
import jax
from jax import lax
import jax.numpy as jnp
import numpy as onp


@flax.struct.dataclass
class Context:
  key: Optional[jax.Array]
  train_step: Optional[int]


def _context_split(context: Context) -> tuple[Context, Context]:
  def mk_ctx(key):
    return Context(key=key, train_step=context.train_step)

  if context.key is not None:
    key1, key2 = jax.random.split(context.key)
    return mk_ctx(key1), mk_ctx(key2)
  return mk_ctx(None), mk_ctx(None)


def _scale_quant(x, *, cfg, ca, context):
  """The core quantizing function."""
  msg = (
      'use_fake_quant mode is used in tests and it is exactly equal when'
      ' po2_scale == True; Did you forget to set it?'
  )
  assert (not cfg.use_fake_quant) or cfg.po2_scale, msg

  # TODO(lew): We should cast earlier. xhs_q should be in cfg.xhs.dtype
  # TODO(lew): After we implement optimization to not double-quantize,
  #   what would happen if we pass fq value (xhs_q2) in residual?

  if isinstance(cfg.numerics, config.NoNumerics):
    return x, None, None
  shared_axes = cfg.calib_shared_axes or ca
  bound = cfg.calibration.get_bound(x, shared_axes)
  abs_max_mapped_to = cfg.numerics.abs_val_mapped_to()
  scale = abs_max_mapped_to / bound

  if cfg.po2_scale:
    # With floor the biggest value (we are using jnp.max) is in the range of
    # clipping and therefore have a correct gradinet.
    scale = 2 ** jnp.floor(jnp.log2(scale))
  if cfg.scale_stop_grad:
    # TODO(lew): Does not matter in DG, because we are using custom gradient.
    #   We should take that into account somehow.
    scale = lax.stop_gradient(scale)

  x_s = _maybe_mul(x, scale)

  quant = jax.custom_vjp(cfg.numerics.fwd)
  quant.defvjp(cfg.numerics.vjp_fwd, cfg.numerics.vjp_bwd)
  quant = functools.partial(quant, context=context)

  x_q, quant_grad = jax.vjp(quant, x_s)
  # We are passing quant_grad (and not more) ot the backward pass.
  # That is equivalent to having:
  # scale = stop_gradient(scale)
  #
  # This is not the only possible choice and we intend to allow experimentation.
  # However for today we hardcoded this choice.
  #
  # In order to achevie no-stop-gradiend solution, we should take vjp
  # of a larger piece of code like the whole _scale_quant.
  #
  # TODO(lew): Implement configuration of stop-gradient.
  inv_scale = _maybe_inv(scale)

  return x_q, inv_scale, quant_grad


def make_fake_quant(cfg: config.Tensor, ca=None):
  def fake_quant(x, context):
    x_q, inv_scale, _ = _scale_quant(x, cfg=cfg, ca=ca, context=context)
    return _maybe_mul(x_q, inv_scale)

  return fake_quant


@flax.struct.dataclass
# It is used only when use_fwd_quant = True
class QTensor:
  qvalue: jnp.ndarray
  qvalue_scale_t: jnp.ndarray


@flax.struct.dataclass
class MultiTensor:
  x: jnp.ndarray
  qx: Optional[QTensor]


@flax.struct.dataclass
class TensorRes:
  """All the things we pass from the forward pass to the backward pass."""
  mt: MultiTensor
  quant_grad: Union[Callable[[jnp.ndarray], tuple[jnp.ndarray]], None]


@flax.struct.dataclass
class DotGeneralRes:
  context_bwd: Context
  lhs: TensorRes
  rhs: TensorRes


def _scale_trans(x, ca, ba):
  """Transposes x to output dimension order."""
  ca = list(ca)
  ba = list(ba)
  for i in ca:
    assert x.shape[i] == 1
  ra = list(i for i in range(len(x.shape)) if i not in ba + ca)
  x = jnp.transpose(x, ba + ra + ca)
  # TODO(lew): x = jnp.squeeze(x, axis=range(len(ba+ra): len(x.shape))
  shape_ba = x.shape[: len(ba)]
  shape_ra = x.shape[len(ba) : len(x.shape) - len(ca)]
  # Will need to add additional axes (size 1) for the other shape_ra
  x = x.reshape(shape_ba + shape_ra)
  return x


def _lhs_scale_transpose(lhs_scale, dimension_numbers, lhs_shape, rhs_shape):
  """Transposes lhs_scale to output dimension order."""
  if lhs_scale is None:
    return None
  # The axis order in out is as follows: batch, lhs_ra, rhs_ra
  # - batch axes order is uniquely determined by either lhs_ba or rhs_ba
  # - contraction axes ca disappear from the output
  # - order of the remaining axes (ra) is preserved.
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  qlhs_scale_t = _scale_trans(lhs_scale, lhs_ca, lhs_ba)
  # inserting dummy axes for rhs_ra
  assert len(qlhs_scale_t.shape) == len(lhs_shape) - len(lhs_ca)
  start = len(qlhs_scale_t.shape)
  end = len(rhs_shape) - len(rhs_ca) - len(rhs_ba) + start
  lhs_dummy_axes = range(start, end)
  qlhs_scale_t = jnp.expand_dims(qlhs_scale_t, axis=lhs_dummy_axes)
  return qlhs_scale_t


def _rhs_scale_transpose(rhs_scale, dimension_numbers, lhs_shape, rhs_shape):
  if rhs_scale is None:
    return None
  del rhs_shape
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  qrhs_scale_t = _scale_trans(rhs_scale, rhs_ca, rhs_ba)
  start = len(rhs_ba)
  end = len(lhs_shape) - len(lhs_ca) - len(lhs_ba) + start
  rhs_dummy_axes = range(start, end)
  qrhs_scale_t = jnp.expand_dims(qrhs_scale_t, axis=rhs_dummy_axes)
  return qrhs_scale_t


def _maybe_mul(x, scale):
  if scale is None:
    return x
  return x * scale


def _maybe_inv(x):
  if x is None:
    return None
  return 1.0 / x


def _make_dot_general_raw(gcfg: config.DotGeneralRaw):
  """Makes quantized lax.dot_general replacement."""

  msg = 'Custom calib_shared_axes not implemented for local AQT.'
  assert gcfg.lhs.calib_shared_axes is None, msg
  assert gcfg.rhs.calib_shared_axes is None, msg

  def dot_general_raw(
      lhs: jnp.ndarray,
      rhs: Union[jnp.ndarray, MultiTensor],
      dimension_numbers,
      context,
  ):
    """Creates a dot_general function without custom gradient."""
    (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
    # We need to copy because we modify cfg to populate some defaults.
    cfg = copy.deepcopy(gcfg)

    # TODO(lew):
    #  - Use qx.value with the int type.
    #  - Handle qx.value with the int type in an optimized way.
    #  - Add a "FQ" case we multiply qx.value*qx.value_scale (not transposed).
    #  - Can we carry untransposed scale and transpose here?
    if isinstance(rhs, MultiTensor):
      # We are in gradient code.
      fwd_quantized = rhs.qx is not None
      expect_fwd_quantized = cfg.rhs.use_fwd_quant is not None
      msg = (
          'Misconfiguration: use_fwd_quant=True, but there is no fwd'
          ' quantization (but rhs.qx is None).'
      )
      assert fwd_quantized == expect_fwd_quantized, msg
      if cfg.rhs.use_fwd_quant:
        assert rhs.qx is not None, msg
        lhs = _maybe_mul(lhs, rhs.qx.qvalue_scale_t)
        rhs = rhs.qx.qvalue
      else:
        rhs = rhs.x
    else:
      assert cfg.rhs.use_fwd_quant is None, 'cannot set use_fwd_quant in fwd'

    if cfg.local_aqt is not None:

      def factor_reshape(x, ca, ba):
        factor = cfg.local_aqt.contraction_axis_shard_count
        assert factor is not None
        if len(ca) == 0:
          return x, ca, ba
        shape = list(x.shape)
        ax = ca[0]
        orig_size = shape[ax]
        assert orig_size % factor == 0
        shape[ax] = factor
        shape.insert(ax + 1, orig_size // factor)
        new_ca = [(b + int(b >= ax)) for b in ca]
        assert new_ca[0] == ax + 1
        new_ba = [ax] + [(b + int(b > ax)) for b in ba]
        return x.reshape(shape), new_ca, new_ba

      lhs, lhs_ca, lhs_ba = factor_reshape(lhs, lhs_ca, lhs_ba)
      rhs, rhs_ca, rhs_ba = factor_reshape(rhs, rhs_ca, rhs_ba)

      dimension_numbers = (lhs_ca, rhs_ca), (lhs_ba, rhs_ba)

    assert isinstance(rhs, jnp.ndarray)

    context, context_bwd = _context_split(context)
    context_lhs, context_rhs = _context_split(context)
    del context

    lhs_q, lhs_inv_scale, lhs_quant_grad = _scale_quant(
        lhs, cfg=cfg.lhs, ca=lhs_ca, context=context_lhs
    )
    lhs_inv_scale_t = _lhs_scale_transpose(
        lhs_inv_scale, dimension_numbers, lhs.shape, rhs.shape
    )
    lhs_qx = (
        None
        if lhs_inv_scale_t is None
        else QTensor(qvalue=lhs_q, qvalue_scale_t=lhs_inv_scale_t)
    )
    lhs_mt = MultiTensor(x=lhs, qx=lhs_qx)
    lhs_res = TensorRes(mt=lhs_mt, quant_grad=lhs_quant_grad)

    rhs_q, rhs_inv_scale, rhs_quant_grad = _scale_quant(
        rhs, cfg=cfg.rhs, ca=rhs_ca, context=context_rhs
    )
    rhs_inv_scale_t = _rhs_scale_transpose(
        rhs_inv_scale, dimension_numbers, lhs.shape, rhs.shape
    )
    rhs_qx = (
        None
        if rhs_inv_scale_t is None
        else QTensor(qvalue=rhs_q, qvalue_scale_t=rhs_inv_scale_t)
    )
    rhs_mt = MultiTensor(x=rhs, qx=rhs_qx)
    rhs_res = TensorRes(mt=rhs_mt, quant_grad=rhs_quant_grad)

    # TODO(lew): mt.x above should be clipped for clipping calibrations

    # These types match default TPU behavior. GPU would need some work.
    # Relevant: https://github.com/google/jax/issues/14022
    # We need this assertion, because we are using lhs.dtype as out dtype.
    assert lhs.dtype == rhs.dtype

    if cfg.lhs.use_fake_quant:
      msg = "Can't set dg_in_dtype in fake_quant mode."
      assert cfg.dg_in_dtype is None, msg
      lhs_q = _maybe_mul(lhs_q, lhs_inv_scale)
      rhs_q = _maybe_mul(rhs_q, rhs_inv_scale)
    else:
      if cfg.dg_in_dtype is not None:
        lhs_q = lhs_q.astype(cfg.dg_in_dtype)
        rhs_q = rhs_q.astype(cfg.dg_in_dtype)

    if cfg.lhs.preprocess_quant_cls is not None:
      preprocess_quant_lhs = cfg.lhs.preprocess_quant_cls()
      lhs_q = preprocess_quant_lhs(lhs_q)
    if cfg.lhs.preprocess_scale_cls is not None:
      preprocess_scale_lhs = cfg.lhs.preprocess_scale_cls()
      lhs_inv_scale_t = preprocess_scale_lhs(lhs_inv_scale_t)
    if cfg.rhs.preprocess_quant_cls is not None:
      preprocess_quant_rhs = cfg.rhs.preprocess_quant_cls()
      rhs_q = preprocess_quant_rhs(rhs_q)
    if cfg.rhs.preprocess_scale_cls is not None:
      preprocess_scale_rhs = cfg.rhs.preprocess_scale_cls()
      rhs_inv_scale_t = preprocess_scale_rhs(rhs_inv_scale_t)
    out = lax.dot_general(
        lhs_q,
        rhs_q,
        dimension_numbers=dimension_numbers,
        preferred_element_type=cfg.dg_accumulator_dtype,
        precision=lax.Precision.DEFAULT,
    ).astype(lhs.dtype)

    if not cfg.lhs.use_fake_quant:
      out = _maybe_mul(out, lhs_inv_scale_t)
      out = _maybe_mul(out, rhs_inv_scale_t)

    res = DotGeneralRes(
        context_bwd=context_bwd,
        lhs=lhs_res,
        rhs=rhs_res,
    )
    if cfg.local_aqt is not None:
      assert len(lhs_ca) == len(rhs_ca)
      if len(lhs_ca) > 0:
        out = jnp.sum(out, axis=0)
      # We are not supporting local AQT in fwd pass, so no res needed.
      res = None
    return out, res

  return dot_general_raw


def _dot_general_raw_attach_gradient(
    fwd_dot_general_raw,
    dlhs_dot_general_raw,
    drhs_dot_general_raw,
):
  """Makes quantized lax.dot_general replacement with attached gradients."""

  def make_fwd(return_residual):
    def fwd(
        lhs,
        rhs,
        dimension_numbers,
        context,
    ):
      assert lhs.dtype == rhs.dtype
      ret, res = fwd_dot_general_raw(
          lhs,
          rhs,
          dimension_numbers,
          context,
      )
      ret = ret.astype(lhs.dtype)
      return (ret, res) if return_residual else ret

    return fwd

  def vjp_bwd(
      fwd_dimension_numbers,
      res: DotGeneralRes,
      g,
  ):
    def ranges_like(*xs):
      start = 0
      for x in xs:
        yield tuple(range(start, start + len(x)))
        start += len(x)

    def grad_dot_general(
        y_res: TensorRes,
        quant_grad,
        dot_general,
        y_is_lhs,
        context,
    ):
      y_ndim = y_res.mt.x.ndim

      (x_ca, y_ca), (x_ba, y_ba) = fwd_dimension_numbers
      if y_is_lhs:
        (y_ca, x_ca) = (x_ca, y_ca)
        (y_ba, x_ba) = (x_ba, y_ba)
      g_ndim = g.ndim - y_ndim + len(x_ba) + 2 * len(x_ca)
      x_ra = tuple(i for i in range(g_ndim) if i not in x_ca and i not in x_ba)
      y_ra = tuple(i for i in range(y_ndim) if i not in y_ca and i not in y_ba)
      if y_is_lhs:
        g_ba, g_ca, _ = ranges_like(x_ba, y_ra, x_ra)
      else:
        g_ba, _, g_ca = ranges_like(x_ba, x_ra, y_ra)
      dims = ((g_ca, y_ra), (g_ba, y_ba))

      out, _ = dot_general(g, y_res.mt, dims, context)

      x_ca_sorted_by_y = tuple(onp.take(x_ca, onp.argsort(y_ca)))
      out_axes = tuple(onp.argsort(tuple(x_ba) + x_ra + x_ca_sorted_by_y))
      transposed_out = jax.lax.transpose(out, out_axes)
      if quant_grad is not None:
        transposed_out = quant_grad(transposed_out)[0]
      return transposed_out

    context1, context2 = _context_split(res.context_bwd)
    dlhs = grad_dot_general(
        res.rhs,
        res.lhs.quant_grad,
        dlhs_dot_general_raw,
        False,
        context1,
    )
    drhs = grad_dot_general(
        res.lhs,
        res.rhs.quant_grad,
        drhs_dot_general_raw,
        True,
        context2,
    )
    return (dlhs, drhs, None)

  vjp = jax.custom_vjp(make_fwd(False), nondiff_argnums=(2,))
  vjp.defvjp(make_fwd(True), vjp_bwd)
  return vjp


def make_dot_general(cfg: Optional[config.DotGeneral]):
  """Makes quantized lax.dot_general replacement with attached gradients."""
  if cfg is None:
    def ret_lax_dg(
        lhs,
        rhs,
        dimension_numbers,
        precision=None,
        preferred_element_type=None,
        *,
        context=Context(key=None, train_step=None),
    ):
      del context
      return jax.lax.dot_general(
          lhs, rhs, dimension_numbers, precision, preferred_element_type
      )

    return ret_lax_dg

  dg = _dot_general_raw_attach_gradient(
      fwd_dot_general_raw=_make_dot_general_raw(cfg.fwd),
      dlhs_dot_general_raw=_make_dot_general_raw(cfg.dlhs),
      drhs_dot_general_raw=_make_dot_general_raw(cfg.drhs),
  )

  def ret_dg(
      lhs,
      rhs,
      dimension_numbers,
      precision=None,
      preferred_element_type=None,
      *,
      context=Context(key=None, train_step=None),
  ):
    del preferred_element_type
    assert (
        precision is None
    ), f'Precision {precision} requested together with quantization.'
    assert lhs.dtype == rhs.dtype, (
        'The only reason we need that, is because we need to determine return'
        ' type.'
    )
    out = dg(
        lhs=lhs,
        rhs=rhs,
        dimension_numbers=dimension_numbers,
        context=context,
    )
    return out

  return ret_dg


def make_conv_general_dilated(cfg: config.DotGeneralRaw):
  """Makes quantized lax.make_conv_general_dilated replacement."""
  # TODO(lew): Either rename DotGeneralConfig or make a conv-specific cfg.
  cfg = copy.deepcopy(cfg)
  if cfg is None:
    cfg = config.DotGeneralRaw.make()

  def my_conv_general_dilated(
      lhs,
      rhs,
      window_strides,
      padding,
      lhs_dilation=None,
      rhs_dilation=None,
      dimension_numbers=None,
      feature_group_count=1,
      batch_group_count=1,
      precision=None,
      preferred_element_type=None,
  ) -> jax.Array:
    msg1 = """
To simplify the code, we currently assume a Flax-particular layout of the data.
This makes sense, because this is the main use-case of this function.
However if there is any other use, we will drop that assumption."""
    rank = len(lhs.shape)
    assert len(rhs.shape) == rank
    assert dimension_numbers is not None, msg1
    assert dimension_numbers.lhs_spec[0:2] == (0, rank - 1), msg1
    assert dimension_numbers.rhs_spec[0:2] == (rank - 1, rank - 2), msg1
    assert dimension_numbers.out_spec[0:2] == (0, rank - 1), msg1
    # In Flax, lhs is the inputs, rhs is the kernel.
    # lhs layout is B, spatials..., Ci
    # rhs layout is: spatials..., Ci, Co
    # out layous it: B, spatials..., Co
    #
    # we need to share these axes: lhs[1:] , rhs[:-1]
    # we have a scale/invscale per: lhs[0] / out[0] and rhs[-1] / out[-1]

    # Flax assumptions.
    assert cfg.lhs.calib_shared_axes == list(range(1, rank))
    assert cfg.rhs.calib_shared_axes == list(range(0, rank - 1))

    lhs_q, lhs_inv_scale, _ = _scale_quant(
        lhs, cfg=cfg.lhs, ca=None, context=None
    )
    rhs_q, rhs_inv_scale, _ = _scale_quant(
        rhs, cfg=cfg.rhs, ca=None, context=None
    )

    out = lax.conv_general_dilated(
        lhs=lhs_q,
        rhs=rhs_q,
        window_strides=window_strides,
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
        precision=precision,
        preferred_element_type=preferred_element_type,
    )

    out = _maybe_mul(out, lhs_inv_scale)
    out = _maybe_mul(out, rhs_inv_scale)

    # # Future scale granularity optimization.
    # In 1x1 conv, each pixel (spatial location) can have different scales
    # in 1xN (rows x colums) conv each row can have different scale, but
    # columns need to share the scales ,  because we are adding pixels across.
    #
    # For patch convs we could have separate scales per patch.
    # We don't do that optimization, because there is a  Flax op: ConvLocal
    # using lax.conv_general_dilated_local which uses lax.dot_general.
    #
    # Dilations: If a dilation of LHS is bigger than the total spatial size of
    # RHS, we could use separe (per LHS pixel) scales.
    # The same applies to dilated RHS.
    # We don't do that optimization yet.
    #
    # We can have different scales across different groups.
    # This applies to both feature and batch.
    return out

  return my_conv_general_dilated



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for matmul."""

import copy
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
from aqt.common import aqt_config
from aqt.tensorflow import aqt_matmul
from aqt.tensorflow import aqt_tensor
from aqt.test import aqt_matmul_test_base
import numpy as np
import tensorflow.compat.v1 as tf


def update_event_count(matmul, event_count_int: int):
  """Update the quantizer's event count without changing stats."""
  for quantizer in [matmul.lhs_quantizer, matmul.rhs_quantizer]:
    sample = tf.zeros(quantizer.data_shape)
    weights = tf.zeros([1] * len(quantizer.data_shape))
    event_count = tf.constant(event_count_int, tf.int64)
    quantizer.update(sample, weights, event_count).run()


def matmul_config(matmul):
  """Creates an AqtMatmulConfig corresponding to a Matmul."""
  return aqt_config.AqtMatmulConfig(matmul.lhs_quantizer.config,
                                    matmul.rhs_quantizer.config)


def _generate_grad_settings():
  scale = 10.0

  float_config_tc = aqt_config.AqtTensorConfig(
      freeze_scale_at_begin=True,
      quant_config=aqt_config.FloatConfig(),
      calibration_config=aqt_matmul_test_base.calibration_config(1))
  float_config = aqt_config.AqtScheduleConfig(
      aqt_matmul_test_base.test_stats_config(), [float_config_tc])
  int_config = aqt_matmul_test_base._schedule_config(8, scale, (0, 1))

  return (
      # Quantization in the forward direction only.
      # In this case we expect the relative error to be in the range
      # [0.01, 1.00] due to the int8 quantization of the LHS.
      dict(
          testcase_name="no_grad_quantization",
          config=aqt_config.AqtMatmulConfig(
              lhs=int_config, rhs=float_config, grad=None),
          check_individual_entries=True,
          minimum_rel_error=0.01,
          maximum_rel_error=1.0,
          scale=scale,
      ),
      # Grad quantization here should not degrade the error by much.
      # In addition to the quantization of the LHS, we also quantize the
      # gradient. But since the gradient has 16 bits we don't expect it to
      # impact the relative error by much.
      dict(
          testcase_name="16_bit_grad_quantization",
          config=aqt_config.AqtMatmulConfig(
              lhs=int_config,
              rhs=float_config,
              grad=aqt_matmul_test_base._schedule_config(
                  bits=16,
                  const_bound_coeff=100,
                  share_stats_axes=(0, 1),
                  freeze_scale_at_begin=False)),
          check_individual_entries=True,
          minimum_rel_error=0.01,
          maximum_rel_error=1.0,
          scale=scale,
      ),
      # Using low-bit quantization will degrade the error.
      # In addition to the quantization of the LHS, we apply a very coarse 1-bit
      # quantization to the gradient. Here we expect the error to be very bad.
      # In particular, we check that the minimum relative error is at least 1.0,
      # which was the upper bound of the cases above.
      dict(
          testcase_name="1_bit_grad_quantization",
          config=aqt_config.AqtMatmulConfig(
              lhs=int_config,
              rhs=float_config,
              grad=aqt_matmul_test_base._schedule_config(
                  bits=1,
                  const_bound_coeff=100,
                  share_stats_axes=(0, 1),
                  freeze_scale_at_begin=False)),
          check_individual_entries=False,
          minimum_rel_error=1.0,
          maximum_rel_error=None,
          scale=scale,
      ),
      # Quantize only the gradient.
      # We don't quantize the LHS, and the gradient is quantized using 16 bits.
      # So, the error we expect should be quite small. In particular, we can
      # upper bound it by 0.01, which was the lower bound for the
      # no_grad_quantization and 16_bit_grad_quantization cases above.
      dict(
          testcase_name="16_bit_grad_quantization_only",
          config=aqt_config.AqtMatmulConfig(
              lhs=float_config,
              rhs=float_config,
              grad=aqt_matmul_test_base._schedule_config(
                  bits=16,
                  const_bound_coeff=100,
                  share_stats_axes=(0, 1),
                  freeze_scale_at_begin=False)),
          check_individual_entries=True,
          minimum_rel_error=1e-5,
          maximum_rel_error=0.01,
          scale=scale,
      ),)


class IntNarrowedMatMulTest(tf.test.TestCase, parameterized.TestCase):

  def test_chooses_right_matmul(self):
    # Create a list of settings (left_bits, right_bits, expected_matmul)
    # and generate a schedule based on those settings.
    settings = [(8, 16, "default"), (4, 16, "default"), (4, 8, "int8"),
                (8, 9, "default"), (4, 4, "int8"), (3, 3, "int8"),
                (9, 7, "default")]

    lhs_schedule = []
    rhs_schedule = []
    expected_results = []
    for i, (l, r, expected) in enumerate(settings):
      lhs_schedule.append((i, i + 1, l))
      rhs_schedule.append((i, i + 1, r))
      expected_results.append(expected)

    lhs_config = aqt_matmul_test_base.config_from_schedule(lhs_schedule)
    rhs_config = aqt_matmul_test_base.config_from_schedule(rhs_schedule)

    shape = [1, 1]  # Any shape will do, we're mocking.
    lhs_quant = aqt_tensor.TensorQuantizer(shape, lhs_config, name="lhs")
    rhs_quant = aqt_tensor.TensorQuantizer(shape, rhs_config, name="rhs")

    module = "aqt.tensorflow.aqt_matmul"
    with mock.patch(f"{module}.default_matmul") as default_matmul, \
         mock.patch(f"{module}.int8_matmul") as int8_matmul:
      default_matmul.return_value = tf.constant("default")
      int8_matmul.return_value = tf.constant("int8")

      event_ph = tf.placeholder(tf.int64)
      lhs_quant._last_update = event_ph
      rhs_quant._last_update = event_ph
      tf_actual = aqt_matmul._matmul_case(lhs_quant, rhs_quant, None, None,
                                          True)

      with self.cached_session():
        tf.global_variables_initializer().run()
        for i, expected in enumerate(expected_results):
          actual = tf_actual.eval(feed_dict={event_ph: i})
          self.assertEqual(
              actual.decode("utf-8"), expected, msg=f"event_count {i}")


class MatmulTest(aqt_matmul_test_base.MatmulTest):

  def constant(self, x):
    return tf.constant(x, dtype=tf.float32)

  def matmul(self, config, lhs_shape, rhs_shape, name="aqt"):
    return aqt_matmul.Matmul(config, lhs_shape, rhs_shape, name)

  def matmul_apply(self, mm, lhs, rhs, train=True, keep_stats=False):
    event_count = tf.constant(0, tf.int64)
    lhs_sample = tf.zeros_like(lhs) if keep_stats else lhs
    lhs_weight = tf.ones_like(lhs) if keep_stats else None
    rhs_sample = tf.zeros_like(rhs) if keep_stats else rhs
    rhs_weight = tf.ones_like(rhs) if keep_stats else None
    updates = [
        mm.update_lhs(lhs_sample, lhs_weight, event_count),
        mm.update_rhs(rhs_sample, rhs_weight, event_count)
    ]
    with tf.control_dependencies(updates):
      result = mm.apply(lhs, rhs, train=train)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      return result.eval()

  def matmul_unquantized(self, lhs, rhs):
    result = tf.matmul(lhs, rhs)
    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      return result.eval()

  def gradients(self, fwd_func, x, w, use_reduce=False):
    if use_reduce:
      fwd_func = lambda x, w: tf.reduce_sum(fwd_func(x, w)**2)
    fwd = fwd_func(x, w)
    return tf.gradients([fwd], [x, w])

  def with_config(self, mm, config):
    """Returns new Matmul with the new config but otherwise the same."""
    with tf.variable_scope(None, default_name="uniqued"):
      return aqt_matmul.Matmul(config, mm.lhs_quantizer.data_shape,
                               mm.rhs_quantizer.data_shape, mm.name,
                               mm.lhs_name, mm.rhs_name)

  def test_validates_contraction(self):
    mm, _, _ = self.exact_int8_matmul_example()

    config = copy.deepcopy(matmul_config(mm))
    config.rhs.stats_config.share_stats_axes = [1]
    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "expected rhs matmul contraction axis"):
      self.with_config(mm, config)

    config = copy.deepcopy(matmul_config(mm))
    config.lhs.stats_config.share_stats_axes = [0]
    with self.assertRaisesRegex(aqt_config.ConfigError,
                                "expected lhs matmul contraction axis"):
      self.with_config(mm, config)

  def test_validates_rank2(self):
    mm, lhs, rhs = self.exact_int8_matmul_example()

    mm.rhs_quantizer.data_shape.append(1)
    with self.assertRaisesRegex(aqt_config.ConfigError, "rhs data shape"):
      mm.apply(lhs, rhs)
    mm.rhs_quantizer.data_shape = mm.rhs_quantizer.data_shape[:-1]

    mm.lhs_quantizer.data_shape += (1,)
    with self.assertRaisesRegex(aqt_config.ConfigError, "lhs data shape"):
      mm.apply(lhs, rhs)

  @parameterized.named_parameters(*_generate_grad_settings())
  def test_grad_linearity(
      self,
      config,
      check_individual_entries,
      minimum_rel_error,
      maximum_rel_error,
      scale,
  ):
    """Validates gradients are correct on basic example."""
    contract_dim = 10
    lhs_shape = (1, contract_dim)
    rhs_shape = (contract_dim, 1)
    target_shape = lhs_shape[:1] + rhs_shape[1:]

    lhs_ph = tf.placeholder(tf.float32, shape=lhs_shape)
    rhs_ph = tf.placeholder(tf.float32, shape=rhs_shape)
    target_ph = tf.placeholder(tf.float32, shape=target_shape)

    mm = aqt_matmul.Matmul(config, lhs_shape, rhs_shape)

    with self.cached_session() as sess, sess.as_default():
      tf.global_variables_initializer().run()
      if config.grad:
        with self.subTest("Gradient shape"):
          assert mm.grad_quantizer is not None
          self.assertSequenceEqual(mm.grad_quantizer.data_shape, target_shape)

      event_count = tf.constant(0, tf.int64)
      updates = [
          mm.update_lhs(tf.ones(lhs_shape), None, event_count),
          mm.update_rhs(tf.ones(rhs_shape), None, event_count),
      ]
      with tf.control_dependencies(updates):
        aqt_mm = mm.apply(lhs_ph, rhs_ph)

      aqt_diff = aqt_mm - target_ph
      aqt_loss = tf.reduce_sum(aqt_diff**2) / 2
      aqt_mm_grad = tf.gradients([aqt_loss], [rhs_ph])[0]

      rng = np.random.default_rng(1234)
      for i in range(10):
        lhs = rng.standard_normal(lhs_shape).astype(np.float32)
        rhs = rng.standard_normal(rhs_shape).astype(np.float32)
        target = rng.standard_normal(target_shape).astype(np.float32)

        feed_dict = {lhs_ph: lhs, rhs_ph: rhs, target_ph: target}

        aqtd, aqt_grad = sess.run([aqt_diff, aqt_mm_grad], feed_dict=feed_dict)

        # Notice aqt gradient at position i is quantized(lhs)[i] * aqtd
        # assuming linearity of gradients.
        grad_factor = aqtd.ravel()
        float_grad = lhs.ravel() * grad_factor
        true_grad = aqt_grad.ravel()

        diff = np.abs(float_grad - true_grad)
        bucket_width = scale * 2 / 255
        for j, err in enumerate(diff):
          if check_individual_entries:
            with self.subTest("Individual entry errors"):
              self.assertLessEqual(
                  err,
                  bucket_width * abs(grad_factor),
                  msg=f"trial {i} position {j}")
        if minimum_rel_error:
          with self.subTest("Minimum overall error"):
            self.assertGreaterEqual(
                np.abs(np.sum(diff) / np.sum(float_grad)), minimum_rel_error)
        if maximum_rel_error:
          with self.subTest("Maximum overall error"):
            self.assertLessEqual(
                np.abs(np.sum(diff) / np.sum(float_grad)), maximum_rel_error)

  @parameterized.named_parameters(
      dict(testcase_name="no_grad", use_grad_quantization=False),
      dict(testcase_name="with_grad", use_grad_quantization=True))
  def test_diagnostics(self, use_grad_quantization):
    mm, lhs, rhs = self.exact_int8_matmul_example(
        use_grad_quantization=use_grad_quantization)

    with self.cached_session():
      tf.global_variables_initializer().run()
      update_event_count(mm, 0)

      # We just want to check if diagnostics works, so use a fake gradient.
      if use_grad_quantization:
        fake_grad = tf.zeros(shape=(lhs.shape[0], rhs.shape[1]))
      else:
        fake_grad = None
      d = mm.diagnostics(lhs, rhs, fake_grad)
      quantizers = {"lhs": mm.lhs_quantizer, "rhs": mm.rhs_quantizer}
      if use_grad_quantization:
        quantizers["grad"] = mm.grad_quantizer

      for qname, quantizer in quantizers.items():
        with self.subTest(f"qname {qname}."):
          for name, expected in quantizer.calibration_variables().items():
            actual = d[f"aqt/{qname}/{name}"]
            self.assertAllEqual(actual, expected)

          actual = d[f"aqt/{qname}/clipped_proportion"]
          expected = 0.0
          self.assertAllEqual(actual, expected)

          actual = d[f"aqt/{qname}/clip"]
          expected = quantizer.clip_range()
          self.assertAllEqual(actual, expected)

      with self.subTest("Out of range"):
        out_of_range_lhs, out_of_range_rhs = (
            tf.ones_like(x) * 512.0 for x in (lhs, rhs))
        d = mm.diagnostics(out_of_range_lhs, out_of_range_rhs)
        for arg in ["lhs", "rhs"]:
          actual = d[f"aqt/{arg}/clipped_proportion"]
          expected = 1.0
          self.assertAllEqual(actual, expected)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration dataclasses."""

import abc
import dataclasses
from typing import Any, Callable, Optional, Union
from aqt.jax.v2 import calibration
from aqt.jax.v2 import int_numerics
from aqt.jax.v2 import stochastic_rounding
import flax.struct
import jax
import jax.numpy as jnp

DType = Any
Context = Any  # TODO(lew): We could put Context in a separate file.

ClipAndRoundFn = Callable[[jnp.ndarray, Context], jnp.ndarray]


class Preprocess(abc.ABC):

  @abc.abstractmethod
  def __call__(self, inputs):
    pass


class NoNumerics(flax.struct.PyTreeNode):
  # TODO(lew): This is a workaround. We should separate Stochastic Rounding.
  # noise_fn has no effect in NoNumerics.
  noise_fn: Optional[stochastic_rounding.NoiseFn] = None

  """No quantization, use a native type such as bf16."""
  pass


Numerics = Union[NoNumerics, int_numerics.IntNumerics]


@dataclasses.dataclass
class Tensor:
  """Configuration of quantization of one tensor or one side of tensor op."""

  numerics: Numerics
  calib_shared_axes: Optional[list[int]]
  scale_stop_grad: bool
  # noise+clip+round
  # We apply gradient of clip_and_round in bwd pass.
  calibration: calibration.Calibration
  # Round up the calibration to power of 2 (po2).
  po2_scale: bool
  use_fake_quant: bool
  # Controls at what value of input tensor should be used.
  # Setting it to True, but not quantizing fwd pass will assert-fail.
  use_fwd_quant: Optional[bool]
  # Operation applied to the quantized inputs to the dot general
  preprocess_quant_cls: Optional[Callable[[], Preprocess]]
  # lax.dot_general output is multiplied by (int transposed) scales.
  # preprocess_scale will be applied to these scales before the multiplication.
  preprocess_scale_cls: Optional[Callable[[], Preprocess]]

  @classmethod
  def make(cls, *args, **kwargs) -> 'Tensor':
    return tensor_make(*args, **kwargs)


@dataclasses.dataclass
class LocalAqt:
  contraction_axis_shard_count: int


@dataclasses.dataclass
class DotGeneralRaw:
  """Configuration of quantization of one dot_general without gradient."""

  lhs: Tensor
  rhs: Tensor
  dg_in_dtype: Optional[DType]
  dg_accumulator_dtype: Optional[DType]
  local_aqt: Optional[LocalAqt]

  @classmethod
  def make(cls, *args, **kwargs) -> 'DotGeneralRaw':
    return dot_general_raw_make(*args, **kwargs)

  @classmethod
  def make_conv_general_dilated(cls, *args, **kwargs) -> 'DotGeneralRaw':
    return conv_general_dilated_make(*args, **kwargs)


@dataclasses.dataclass
class DotGeneral:
  """Configuration of quantization of dot_general and its gradients."""

  fwd: DotGeneralRaw
  dlhs: DotGeneralRaw
  drhs: DotGeneralRaw

  @classmethod
  def make(cls, *args, **kwargs) -> 'DotGeneral':
    return dot_general_make(*args, **kwargs)


################################################################################
# Functions below are auxiliary helpers.


def set_accumulator_dtype(
    cfg: DotGeneral,
    fwd_dtype: Optional[DType],
    dlhs_dtype: Optional[DType],
    drhs_dtype: Optional[DType],
):
  cfg.fwd.dg_accumulator_dtype = fwd_dtype
  cfg.dlhs.dg_accumulator_dtype = dlhs_dtype
  cfg.drhs.dg_accumulator_dtype = drhs_dtype


def set_stochastic_rounding(
    cfg: DotGeneral,
    # Typically we have (but it's a caller's responsibility to check):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    vjp_lhs_stochastic_rounding: bool,
    vjp_rhs_stochastic_rounding: bool,
    implementation: str,
):
  """Configure stochastic rounding implementation."""
  noise_implementations = {
      'jax.uniform': lambda shape, key: jax.random.uniform(key, shape) - 0.5,
      'custom-1': stochastic_rounding.random_centered_uniform,
  }
  msg = f'{implementation} not supported.'
  assert implementation in noise_implementations.keys(), msg
  noise_fn = noise_implementations[implementation]

  if vjp_lhs_stochastic_rounding:
    cfg.dlhs.lhs.numerics = cfg.dlhs.lhs.numerics.replace(noise_fn=noise_fn)
    cfg.drhs.lhs.numerics = cfg.drhs.lhs.numerics.replace(noise_fn=noise_fn)
  else:
    cfg.dlhs.lhs.numerics = cfg.dlhs.lhs.numerics.replace(noise_fn=None)
    cfg.drhs.lhs.numerics = cfg.drhs.lhs.numerics.replace(noise_fn=None)

  if vjp_rhs_stochastic_rounding:
    cfg.dlhs.rhs.numerics = cfg.dlhs.rhs.numerics.replace(noise_fn=noise_fn)
    cfg.drhs.rhs.numerics = cfg.drhs.rhs.numerics.replace(noise_fn=noise_fn)
  else:
    cfg.dlhs.rhs.numerics = cfg.dlhs.rhs.numerics.replace(noise_fn=None)
    cfg.drhs.rhs.numerics = cfg.drhs.rhs.numerics.replace(noise_fn=None)


def set_static_bound(cfg: DotGeneral, bound: float = 1.0):
  cfg.fwd.lhs.calibration = calibration.ConstantCalibration(bound)
  cfg.fwd.rhs.calibration = calibration.ConstantCalibration(bound)
  cfg.drhs.lhs.calibration = calibration.ConstantCalibration(bound)
  cfg.drhs.rhs.calibration = calibration.ConstantCalibration(bound)
  cfg.dlhs.lhs.calibration = calibration.ConstantCalibration(bound)
  cfg.dlhs.rhs.calibration = calibration.ConstantCalibration(bound)


def tensor_make(bits: Optional[int]) -> 'Tensor':
  """Makes config.Tensor."""
  if bits is None:
    numerics = NoNumerics()
  else:
    pz = False if bits == 1 else True
    numerics = int_numerics.IntNumerics(
        bits=bits,
        preserve_zero=pz,
        preserve_max_val=False,
        clip=True,
        round=True,
        noise_fn=None,
        clip_gradient=False,  # This can be disabled when using abs-max scaling.
    )

  return Tensor(
      numerics=numerics,
      calib_shared_axes=None,
      scale_stop_grad=True,
      calibration=calibration.AbsMaxCalibration(),
      po2_scale=False,
      use_fake_quant=False,
      # dtype_x=dtype,
      use_fwd_quant=None,
      preprocess_quant_cls=None,
      preprocess_scale_cls=None,
  )


def dot_general_raw_make(
    lhs_bits=None,
    rhs_bits=None,
    local_aqt=None,
) -> 'DotGeneralRaw':
  """Create quantization configs for input matrices to a matmul."""
  lhs_cfg = tensor_make(lhs_bits)
  rhs_cfg = tensor_make(rhs_bits)

  # Binary uses 0.5 right now.
  if (
      lhs_bits is not None
      and rhs_bits is not None
      and 2 <= lhs_bits <= 8
      and 2 <= rhs_bits <= 8
  ):
    dg_in_dtype = jnp.int8
    dg_accumulator_dtype = jnp.int32
  else:
    # Use None to determine the dtype on the fly in aqt_dot_general
    dg_in_dtype = None
    dg_accumulator_dtype = None

  return DotGeneralRaw(
      lhs=lhs_cfg,
      rhs=rhs_cfg,
      dg_in_dtype=dg_in_dtype,
      dg_accumulator_dtype=dg_accumulator_dtype,
      local_aqt=local_aqt,
  )


def conv_general_dilated_make(
    spatial_dimensions=2,
    lhs_bits: Optional[int] = None,
    rhs_bits: Optional[int] = None,
) -> 'DotGeneralRaw':
  """Create quantization config conv_general_dilated."""
  config = dot_general_raw_make(lhs_bits, rhs_bits)
  # Hardcoding flax assumptions.
  if config.lhs:
    config.lhs.calib_shared_axes = list(range(1, spatial_dimensions + 2))
  if config.rhs:
    config.rhs.calib_shared_axes = list(range(0, spatial_dimensions + 2 - 1))
  return config


def dot_general_make(
    lhs_bits: Optional[int] = None,
    rhs_bits: Optional[int] = None,
    bwd_bits: Optional[int] = None,
    use_fwd_quant: bool = True,
    dlhs_local_aqt=None,
    drhs_local_aqt=None,
) -> 'DotGeneral':
  """Create quantization configs for input matrices to a matmul."""
  fwd = dot_general_raw_make(lhs_bits, rhs_bits)
  dlhs = dot_general_raw_make(bwd_bits, bwd_bits, local_aqt=dlhs_local_aqt)
  drhs = dot_general_raw_make(bwd_bits, bwd_bits, local_aqt=drhs_local_aqt)
  cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

  # Surprising: lhs quantization determines what drhs can do.
  if lhs_bits is not None:
    # Only rhs is accepting MultiTensor.
    cfg.drhs.rhs.use_fwd_quant = use_fwd_quant
  if rhs_bits is not None:
    cfg.dlhs.rhs.use_fwd_quant = use_fwd_quant
  return cfg


def fully_quantized(
    *,
    fwd_bits: int | None = 8,
    bwd_bits: int | None = 8,
    use_fwd_quant: bool = True,
    use_stochastic_rounding: Optional[bool] = True,
    # Typically we have (but it's a caller's responsibility to check):
    # - vjp_lhs_stochastic_rounding is referring to the gradient and
    # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
    vjp_lhs_stochastic_rounding: Optional[bool] = None,
    vjp_rhs_stochastic_rounding: Optional[bool] = None,
    # The dummy static bound flag is temporary, for performance benchmarking.
    use_dummy_static_bound: bool = False,
    dlhs_local_aqt: Optional[LocalAqt] = None,
    drhs_local_aqt: Optional[LocalAqt] = None,
) -> DotGeneral:
  """Fully Quantized Training."""
  cfg = dot_general_make(
      lhs_bits=fwd_bits,
      rhs_bits=fwd_bits,
      bwd_bits=bwd_bits,
      use_fwd_quant=use_fwd_quant,
      dlhs_local_aqt=dlhs_local_aqt,
      drhs_local_aqt=drhs_local_aqt,
  )

  # Stochastic Rounding
  # These 3 variables are used to ensure we don't mix
  # old and new style of SR configuration.
  old_style_sr_config = use_stochastic_rounding is not None
  new_style_sr_config_lhs = vjp_lhs_stochastic_rounding is not None
  new_style_sr_config_rhs = vjp_rhs_stochastic_rounding is not None
  assert new_style_sr_config_lhs == new_style_sr_config_rhs, (
      'if you use new style SR config (vjp_xhs_stochastic_rounding), do pass'
      ' both lhs and rhs explicitely.'
  )
  assert new_style_sr_config_lhs != old_style_sr_config

  true = True  # A crude way to get around g-explicit-bool-comparison warning

  assert not (vjp_lhs_stochastic_rounding and vjp_rhs_stochastic_rounding), (
      'This config is buggy when you set both to True. Contact lew@ or use'
      ' config_v3'
  )

  # By default use jax.uniform for stochastic rounding
  if use_stochastic_rounding == true:
    set_stochastic_rounding(cfg, True, True, 'jax.uniform')

  if vjp_lhs_stochastic_rounding == true:
    set_stochastic_rounding(cfg, True, False, 'jax.uniform')

  if vjp_rhs_stochastic_rounding == true:
    set_stochastic_rounding(cfg, False, True, 'jax.uniform')

  if use_dummy_static_bound:
    set_static_bound(cfg, 1.0)

  return cfg


def config_v3(
    *,
    fwd_bits: int | None,
    dlhs_bits: int | None,
    drhs_bits: int | None,
    # The dummy static bound flag is for performance benchmarking.
    use_dummy_static_bound: bool = False,
    rng_type: str = 'jax.uniform',  # 'custom-1'
    dlhs_local_aqt: Optional[LocalAqt] = None,
    drhs_local_aqt: Optional[LocalAqt] = None,
    fwd_accumulator_dtype: ... = jnp.int32,
    dlhs_accumulator_dtype: ... = jnp.int32,
    drhs_accumulator_dtype: ... = jnp.int32,
) -> DotGeneral:
  """Fully Quantized Training."""
  fwd = dot_general_raw_make(fwd_bits, fwd_bits)
  dlhs = dot_general_raw_make(dlhs_bits, dlhs_bits, local_aqt=dlhs_local_aqt)
  drhs = dot_general_raw_make(drhs_bits, drhs_bits, local_aqt=drhs_local_aqt)
  cfg = DotGeneral(fwd=fwd, dlhs=dlhs, drhs=drhs)

  cfg.dlhs.rhs.use_fwd_quant = False
  cfg.drhs.rhs.use_fwd_quant = False

  # Typically we have (but I don't know if it is guraranteed):
  # - vjp_lhs_stochastic_rounding is referring to the gradient and
  # - vjp_rhs_stochastic_rounding is referring to the activations/weights.
  set_stochastic_rounding(
      cfg,
      vjp_lhs_stochastic_rounding=True,
      vjp_rhs_stochastic_rounding=False,
      implementation=rng_type,
  )

  if use_dummy_static_bound:
    set_static_bound(cfg, 1.0)

  set_accumulator_dtype(
      cfg,
      fwd_dtype=fwd_accumulator_dtype,
      dlhs_dtype=dlhs_accumulator_dtype,
      drhs_dtype=drhs_accumulator_dtype,
  )
  return cfg



# Copyright 2022 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized matrix multiplication.

The matmul operation is a form of tensor product applied to two arguments
`(a, b)` which contracts the penultimate axis of `a` with the ultimate axis
of `b`.

For details on quantized operations and common configuration arguments, see
`aqt_ops`.
"""

from typing import Callable, Dict, Iterable, Optional

from aqt.common import aqt_config
from aqt.common import aqt_config_utils
from aqt.tensorflow import aqt_ops_util
from aqt.tensorflow import aqt_tensor
import tensorflow.compat.v1 as tf

# TODO(b/220181240): Remove the pylint disable below and avoid using protected
# methods.
# We repeatedly use protected methods from classes defined in other modules to
# avoid exporting them as part of the public API.
# pylint: disable=protected-access

MatmulFn = Callable[[tf.Tensor, tf.Tensor], tf.Tensor]


def default_matmul(
    lhs_quantizer: aqt_tensor.TensorQuantizer,  #
    rhs_quantizer: aqt_tensor.TensorQuantizer,
    lhs: tf.Tensor,
    rhs: tf.Tensor,
    train: bool,
    transpose_a: bool = False,
    transpose_b: bool = False,
) -> tf.Tensor:
  """Perform tf.matmul with input tensors of float32 type."""
  lhs = aqt_ops_util._possibly_use_quantized_variable(lhs_quantizer, lhs, train)
  rhs = aqt_ops_util._possibly_use_quantized_variable(rhs_quantizer, rhs, train)

  return tf.matmul(lhs, rhs, transpose_a=transpose_a, transpose_b=transpose_b)


def int8_matmul(
    lhs_quantizer: aqt_tensor.TensorQuantizer,  #
    rhs_quantizer: aqt_tensor.TensorQuantizer,
    lhs: tf.Tensor,
    rhs: tf.Tensor,
    train: bool,
    transpose_a: bool = False,
    transpose_b: bool = False,
    ) -> tf.Tensor:
  """Perform integral matmul after i8 cast while preserving gradients.

  1. If {lhs,rhs}_quantizer indicates the currently-active configuration
     should use saved quantized variables, then uses them. Otherwise,
     casts lhs/rhs to tf.int8.
  2. Performs native int8 matmul.
  3. Casts int32 result to f32.

  Despite the effictive clipping, it has a gradient of a float matmul.

  Args:
    lhs_quantizer: TensorQuantizer for lhs.
    rhs_quantizer: TensorQuantizer for rhs.
    lhs: left hand side of the matmul, as a float.
    rhs: right hand side of the matmul, as a float.
    train: If false and `TQ.use_quantized_variable` is True, then use the
      quantized variable, instead of input tensors, for the respective input
      tensor.
    transpose_a: Whether to transpose the LHS.
    transpose_b: Whether to transpose the RHS.

  Returns:
    The result of the integer matmul.
  """

  def grad_fn(dy, variables=None):
    # We could manually write the backward pass of matmul with respect to the
    # inputs lhs, rhs by deriving an equivalent expression with matrix
    # calculus. However, with higher-rank tensors, the axes handling becomes
    # cumbersome. Since the backward pass is in floating point right now anyway,
    # just generate a temporary forward matmul op (the one we quantized in the
    # forward pass) and ask autograd to derive the backwards direction itself.
    mm = tf.matmul(lhs, rhs)
    grads = tf.gradients(
        ys=[mm],
        xs=[lhs, rhs],
        # Stop grads in case inputs are aliased.
        stop_gradients=[lhs, rhs],
        grad_ys=[dy])

    # The variables used by fwd() are quantized, so they have no gradients.
    if variables:
      return grads, [None] * len(variables)
    return grads

  def fwd_no_grad(arg_lhs, arg_rhs):
    # Wrap a custom-gradient op around the cast to propagate gradients,
    # since casting stops the gradient.
    int_lhs = tf.cast(arg_lhs, tf.int8)
    int_rhs = tf.cast(arg_rhs, tf.int8)

    int_lhs = aqt_ops_util._possibly_use_quantized_variable(
        lhs_quantizer, int_lhs, train
    )
    int_rhs = aqt_ops_util._possibly_use_quantized_variable(
        rhs_quantizer, int_rhs, train
    )

    imm = tf.matmul(
        int_lhs,
        int_rhs,
        output_type=tf.int32,
        transpose_a=transpose_a,
        transpose_b=transpose_b,
    )
    mm = tf.cast(imm, tf.float32)
    return mm

  fwd = tf.custom_gradient(lambda l, r: (fwd_no_grad(l, r), grad_fn))

  # If not training, do not install the custom gradient since it would cause
  # both sets of weights (float + int8) to be pulled into the graph, making it
  # difficult to prune away the unused set of weights for serving.
  if train:
    return fwd(lhs, rhs)
  else:
    return fwd_no_grad(lhs, rhs)


def _matmul_case(
    lhs_quantizer,
    rhs_quantizer,
    lhs,
    rhs,
    train,
    transpose_a: bool = False,
    transpose_b: bool = False,
):
  """Switch over matmuls based on event count and configs.

  The `TensorQuantizer`s for each argument are provided to supply the
  configurations for each input tensor. Also, if indicated by constituent
  configs, this uses the quantized variables in each tensor quantizer rather
  than the inputs themselves. Regardless, gradients are preserved as if this was
  a matmul over the original `(lhs, rhs)` inputs.

  Args:
    lhs_quantizer: TensorQuantizer for lhs.
    rhs_quantizer: TensorQuantizer for rhs.
    lhs: float input for lhs.
    rhs: float input for rhs.
    train: If false and `TQ.use_quantized_variable` is True, then use the
      quantized variable, instead of input tensors, for the respective input
      tensor.
    transpose_a: Whether to transpose the LHS.
    transpose_b: Whether to transpose the RHS.

  Returns:
    The `tf.Tensor` from the resulting quantized matmul.
  """

  def cond_int8_matmul():
    return int8_matmul(lhs_quantizer, rhs_quantizer, lhs, rhs, train,
                       transpose_a, transpose_b)

  def cond_default_matmul():
    return default_matmul(lhs_quantizer, rhs_quantizer, lhs, rhs, train,
                          transpose_a, transpose_b)

  return aqt_ops_util._dense_op_case(
      lhs_quantizer, rhs_quantizer, cond_default_matmul, cond_int8_matmul, train
  )


def _validate_inputs(
    lhs_quantizer: aqt_tensor.TensorQuantizer,  #
    rhs_quantizer: aqt_tensor.TensorQuantizer,
    grad_quantizer: Optional[aqt_tensor.TensorQuantizer],
):
  """Validates configs and inputs for matmul."""

  if len(lhs_quantizer.data_shape) != 2:
    raise aqt_config.ConfigError(
        f'lhs data shape ({lhs_quantizer.data_shape}) not rank 2')

  if len(rhs_quantizer.data_shape) != 2:
    raise aqt_config.ConfigError(
        f'rhs data shape ({rhs_quantizer.data_shape}) not rank 2')

  lhs_config = lhs_quantizer.config
  rhs_config = rhs_quantizer.config

  aqt_config_utils._validate_alignment(
      'lhs_config',  #
      lhs_config.tensor_configs,
      'rhs_config',
      rhs_config.tensor_configs)

  if 1 not in lhs_config.stats_config.share_stats_axes:
    raise aqt_config.ConfigError(
        f'expected lhs matmul contraction axis (1) to be in '
        f'share_stats_axes={lhs_config.stats_config.share_stats_axes}')
  if 0 not in rhs_config.stats_config.share_stats_axes:
    raise aqt_config.ConfigError(
        f'expected rhs matmul contraction axis (0) to be in '
        f'share_stats_axes={rhs_config.stats_config.share_stats_axes}')

  if grad_quantizer:
    # Other quantizations are possible -- for now the only quantization we allow
    # for the gradient is based on scalar scaling.
    # For now we also restrict the possible values to dynamic quantization.
    grad_config = grad_quantizer.config
    if (0 not in grad_config.stats_config.share_stats_axes or
        1 not in grad_config.stats_config.share_stats_axes):
      raise aqt_config.ConfigError(
          f'expected grad matmul axes (0, 1) to both be in '
          f'share_stats_axes={rhs_config.stats_config.share_stats_axes}')
    if grad_config.stats_config.ema_update_count != 1:
      raise aqt_config.ConfigError(
          'Currently only supporting dynamic quantization for the gradient ' +
          'with ema_update_count=1, but instead found ' +
          f'{grad_config.stats_config.ema_update_count}.')
    for tensor_config in grad_config.tensor_configs:
      if tensor_config.freeze_scale_at_begin:
        raise aqt_config.ConfigError(
            'Currently only supporting dynamic quantization for the gradient ' +
            'with freeze_scale_at_begin=False, but got True.')


def matmul(
    lhs_quantizer: aqt_tensor.TensorQuantizer,  #
    lhs: tf.Tensor,
    rhs_quantizer: aqt_tensor.TensorQuantizer,
    rhs: tf.Tensor,
    grad_quantizer: Optional[aqt_tensor.TensorQuantizer] = None,
    train: bool = True,
) -> tf.Tensor:
  """Quantized tf.matmul.

  Gradients are propagated using the straight-through estimator [1] to
  `lhs` and `rhs` arguments. Note that this is a pure function, with
  quantization determined by the argument quantizers, which must be
  updated separately.

  If not `lhs_quantizer.train` (which should be equal to that of RHS)
  and the quantizers have cached quantized variables, then those are
  used instead, if the quantization config indicates as much.

  When grad quantization is enabled, we use the approach described here:
  https://g3doc.corp.google.com/third_party/py/aqt/google/g3doc/aqt_math_and_algorithms.md?cl=474412951#one-pragmatic-approach

  Args:
    lhs_quantizer: the tensor quantizer for lhs.
    lhs: left-hand side of the matmul, with shape [B, C].
    rhs_quantizer: the tensor quantizer for rhs.
    rhs: right-hand side of the matmul, with shape [C, F].
    grad_quantizer: Optional quantizer for the gradient, with shape [B, F].
    train: If false and `use_quantized_variable` in lhs_quantizer or
      rhs_quantizer, then this indicates `aqt_matmul` should use the quantized
      variable with the latest quantized, memorized from the most recent
      `TensorQuantizer.update()` in quantized operations rather than the float
      tensor input `lhs` or `rhs` provided to those operations at inference
      time.

  Returns:
    Approximate Matmul result.

  Raises:
    aqt_config.ConfigError: if the configurations for the two quantizers
      are incompatible or unaligned (endpoints for start/stop event counts
      must be the same).

  [1]: https://arxiv.org/abs/1308.3432
  """
  _validate_inputs(lhs_quantizer, rhs_quantizer, grad_quantizer)

  def fwd(lhs, rhs):
    with tf.name_scope('AqtMatMul'):
      with tf.name_scope('get_quant_scale'):
        with tf.name_scope('lhs'):
          lhs_scale, lhs_inv_scale = lhs_quantizer._get_quant_scale(
              train)  # [B, 1]
        with tf.name_scope('rhs'):
          rhs_scale, rhs_inv_scale = rhs_quantizer._get_quant_scale(
              train)  # [1, F]

      lhs = lhs_scale * lhs  # [B, 1] * [B, C]
      rhs = rhs_scale * rhs  # [1, F] * [C, F]

      with tf.name_scope('to_quant'):
        with tf.name_scope('lhs'):
          lhs = lhs_quantizer._to_quant(lhs, train)  # [B, C]
        with tf.name_scope('rhs'):
          rhs = rhs_quantizer._to_quant(rhs, train)  # [C, F]

      with tf.name_scope('matmul'):
        mm = _matmul_case(lhs_quantizer, rhs_quantizer, lhs, rhs,
                          train)  # [B, F]

      with tf.name_scope('inv_scale'):
        # TODO(b/236024344): consider alternative multiply associations here.
        inv_scale = (lhs_inv_scale * rhs_inv_scale)  # [B, 1] * [1, F]
        out = mm * inv_scale  # [B, F] * [B, F]

    return out

  @tf.custom_gradient
  def qmatmul(lhs, rhs):

    out = fwd(lhs, rhs)

    def bwd(grad):
      # Make sure to build all backprop results after fprop is computed.
      # Since we rely on variables being updated, this is important for
      # consistency. For instance, the forward pass might be computed under
      # a user-added control dependency from Matmul.update_{lhs,rhs}; the
      # backward pass should also share this dependency transitively.
      with tf.control_dependencies([out]):

        # Note the differences between autograd and what we
        # have implemented below:
        #
        # (1) We re-quantize lhs, rhs to save memory between
        #     fprop and bprop, this is manual rematerialization.
        # (2) Each bprop matmul can avoid a multiplication
        #     by an scale and inverse scale of an fprop argument
        #     due to cancellation.

        with tf.name_scope('BwdAqtMatMul'):

          with tf.name_scope('get_quant'):
            with tf.name_scope('lhs'):
              lhs_scale, lhs_inv_scale = lhs_quantizer._get_quant_scale(train)
            with tf.name_scope('rhs'):
              rhs_scale, rhs_inv_scale = rhs_quantizer._get_quant_scale(train)

          lhs_scaled = lhs_scale * lhs
          rhs_scaled = rhs_scale * rhs

          if grad_quantizer:
            with tf.name_scope('grad_quant'):
              # [B, F] * [B, 1] * [1, F]
              grad = grad * lhs_inv_scale * rhs_inv_scale
              # For now we simplify the gradient update by disabling the ability
              # to pass the `weight` argument to update.

              # We can also reuse the _last_update from the lhs_quantizer to
              # pass the `event_count` which is expected to be consistent for
              # the LHS, RHS, and gradient.
              op = grad_quantizer.update(
                  grad, weight=None, event_count=lhs_quantizer._last_update)
              with tf.control_dependencies([op]):
                # For now we assume that the gradient quantization is used only
                # at the scalar level, i.e., grad_scale and grad_inv_scale have
                # shape [1, 1]. This is enforced by the validation step.
                # However, in principle the code below does not depend on this
                # assumption. More generally the gradient may have other per
                # channel quantization, in which case the shapes could be
                # [B, 1], [1, F] or [B, F].
                grad_scale, grad_inv_scale = grad_quantizer._get_quant_scale(
                    train)

                grad = grad * grad_scale
                grad = grad_quantizer._to_quant(grad, train)

          with tf.name_scope('lhs'):
            qrhs = rhs_quantizer._to_quant(rhs_scaled, train)
            if grad_quantizer:
              lhs_bwd = grad_inv_scale * lhs_scale * _matmul_case(
                  lhs_quantizer=grad_quantizer,
                  rhs_quantizer=rhs_quantizer,
                  lhs=grad,
                  rhs=qrhs,
                  train=train,
                  transpose_a=False,
                  transpose_b=True)
            else:
              lhs_bwd = tf.matmul(grad * rhs_inv_scale, tf.transpose(qrhs))
            lhs_bwd = tf.where_v2(
                lhs_quantizer._clip_mask(lhs_scaled, train), 0.0, lhs_bwd)

          with tf.name_scope('rhs'):
            qlhs = lhs_quantizer._to_quant(lhs_scaled, train)
            if grad_quantizer:
              rhs_bwd = grad_inv_scale * rhs_scale * _matmul_case(
                  lhs_quantizer=lhs_quantizer,
                  rhs_quantizer=grad_quantizer,
                  lhs=qlhs,
                  rhs=grad,
                  train=train,
                  transpose_a=True,
                  transpose_b=False)
            else:
              rhs_bwd = tf.matmul(tf.transpose(qlhs), grad * lhs_inv_scale)
            rhs_bwd = tf.where_v2(
                rhs_quantizer._clip_mask(rhs_scaled, train), 0.0, rhs_bwd)

        return [lhs_bwd, rhs_bwd]

    return out, bwd

  # If not training, do not install the custom gradient since it would cause
  # both sets of weights (float + int8) to be pulled into the graph, making it
  # difficult to prune away the unused set of weights for serving.
  if train:
    return qmatmul(lhs, rhs)
  else:
    return fwd(lhs, rhs)


class Matmul:
  """Encapsulates a quantized matmul op and calibration state."""

  def __init__(
      self,
      config: aqt_config.AqtMatmulConfig,
      lhs_shape: Iterable[Optional[int]],
      rhs_shape: Iterable[Optional[int]],
      name: str = 'aqt',
      lhs_name: str = 'lhs',
      rhs_name: str = 'rhs',
      grad_name: str = 'grad',
  ):
    """Creates a Matmul instance.

    This encapsulates the state necessary for quantizing both
    of the matmul arguments.

    Args:
      config: an AqtMatmulConfig
      lhs_shape: shape of the lhs tensor to multiply
      rhs_shape: shape of the rhs tensor to multiply
      name: variable scope for all variables to be created.
      lhs_name: scope for left hand side variables only.
      rhs_name: scope for right hand side variables only.
      grad_name: scope for the grad variables only (if quantizing).
    """
    self.name = name
    self.lhs_name = lhs_name
    self.rhs_name = rhs_name
    self.grad_name = grad_name
    with tf.variable_scope(name):
      self.lhs_quantizer = aqt_tensor.TensorQuantizer(
          lhs_shape, config.lhs, name=lhs_name)
      self.rhs_quantizer = aqt_tensor.TensorQuantizer(
          rhs_shape, config.rhs, name=rhs_name)
      if config.grad:  # The gradient shape is inferred from the rhs and lhs.
        grad_shape = [list(lhs_shape)[0], list(rhs_shape)[1]]
        self.grad_quantizer = aqt_tensor.TensorQuantizer(
            grad_shape, config.grad, name=grad_name)
      else:
        self.grad_quantizer = None

    _validate_inputs(self.lhs_quantizer, self.rhs_quantizer,
                     self.grad_quantizer)

  def update_lhs(
      self,
      x: tf.Tensor,
      weights: tf.Tensor,
      event_count: tf.Tensor,
  ) -> tf.Operation:
    """Updates variables for an observation of the lhs.

    Updating variables for an argument updates the statistics
    for a new input to account for incremental observations of
    a tensor's entries' magnitudes.

    This also updates the scales, if they were set according to
    a previous calibration config and now we've moved on to
    a new event associated with a different calibration configuration
    in the schedule.

    Args:
      x: a tensor for the observation of an lhs input
      weights: a weight matrix broadcastable to x, representing how much weight
        the corresponding axes should have on statistics associated with
        quantizing that dimension.
      event_count: the event of the observation

    Returns:
      The tf.Operation corresponding to the update
    """
    return self.lhs_quantizer.update(x, weights, event_count)

  def update_rhs(
      self,
      x: tf.Tensor,
      weights: tf.Tensor,
      event_count: tf.Tensor,
  ) -> tf.Operation:
    """Computes analogue of update_lhs, but for rhs."""
    return self.rhs_quantizer.update(x, weights, event_count)

  def apply(
      self,
      lhs: tf.Tensor,
      rhs: tf.Tensor,
      train: bool = True,
  ) -> tf.Tensor:
    """Generates a pure quantized matmul op.

    Make sure that `apply` is called within the context of any updates
    to statistics used for calibration you'd like to happen before the
    op.

    Args:
      lhs: a float32 tensor for the left hand side
      rhs: a float32 tensor for the right hand side
      train: whether to generate the training or serving graph

    Returns:
      A tf.Tensor generated from possibly quantizing lhs and rhs
      with clip bounds derived from the current quantizer statistics.
    """

    return matmul(
        self.lhs_quantizer,
        lhs,
        self.rhs_quantizer,
        rhs,
        self.grad_quantizer,
        train=train)

  def diagnostics(
      self,
      lhs: tf.Tensor,
      rhs: tf.Tensor,
      grad: Optional[tf.Tensor] = None,
  ) -> Dict[str, tf.Tensor]:
    """Returns a dictionary from keys to diagnostic tensors.

    Args:
      lhs: lhs argument to self.apply, used for deriving diangostics relative to
        a given input.
      rhs: as above, but for rhs
      grad: If specified, the gradient for deriving diagnostics.

    Returns:
      A dictionary with various quantization-related diagnostics,
      whose string keys are prefixed by self.name/self.{lhs,rhs}_name.
    """
    return aqt_ops_util.diagnostics(self, lhs, rhs, grad)
